from ase.neighborlist import neighbor_list
from clease.tools import ApproxEqualityList


class MICDistanceNotUniqueError(Exception):
    pass


class TransMatrixConstructor(object):
    """Class that constructs translation matrices.

    Parameters:

    atoms: Atoms object
        ASE atoms object (assumed to be wrapped and sorted)

    cutoff: float
        Cut-off distance in angstrom
    """

    def __init__(self, atoms, cutoff):
        self.num_atoms = len(atoms)
        self.neighbor = self._construct_neighbor_list(atoms, cutoff)

    def _construct_neighbor_list(self, atoms, cutoff):
        """Construct neighbour list structure."""
        i_first, i_second, d_vec = neighbor_list('ijD', atoms, cutoff)
        # Transfer to more convenienct strucutre
        neighbor = [{"nb_index": [], "dist": []} for _ in range(len(atoms))]

        # Re-group by first index
        for i in range(len(i_first)):
            neighbor[i_first[i]]["nb_index"].append(i_second[i].tolist())
            neighbor[i_first[i]]["dist"].append(
                ApproxEqualityList(d_vec[i].tolist()))

        # Sort based on distance
        for i in range(len(neighbor)):
            srt = sorted(list(zip(neighbor[i]["dist"],
                                  neighbor[i]["nb_index"])))
            srt = list(zip(*srt))  # Unzip the list
            neighbor[i]["dist"] = [l.array for l in srt[0]]
            neighbor[i]["nb_index"] = srt[1]
        return neighbor

    def _map_one(self, indx, template_indx):
        """Map indices of neighbors to another reference atom."""
        mapped = {template_indx: indx}

        nb_indx = self.neighbor[indx]["nb_index"]
        ref_indx = self.neighbor[template_indx]["nb_index"]
        mapped = {ref: ind for ref, ind in zip(ref_indx, nb_indx)}
        mapped[template_indx] = indx
        return mapped

    def construct(self, ref_symm_group, symm_group, indices=None):
        """Construct translation matrix.

        Parameters:

        ref_symm_group: list
            List of reference indices.
            If passed Atoms object has only one basis, ref_symm_group is [0],
            otherwise it hold indices of reference atoms in each basis.
            (e.g., [0, 5, 15] for the Atoms object with 3 basis)

        symm_group: list
            List of the symmetry groups of each Atoms object.
            If passed Atoms object has only one basis, symm_group is
            [0, 0, ..., 0].
            If it has two basis,  this can be [0, 0, 1, 1, 0, 1, ...].
            The index of the reference atom for the basis where the atom
            with index k belongs to is ref_symm_group[symm_group[k]]

        indices: list of int
            List of indices that the translation matrix should be calculated
            for. If None, all indices will be used.
        """
        tm = []
        if indices is None:
            indices = range(self.num_atoms)

        for indx in indices:
            ref_indx = ref_symm_group[symm_group[indx]]
            mapped = self._map_one(indx, ref_indx)
            tm.append(mapped)
        return tm



"""Calculator for Cluster Expansion."""
import sys
import numpy as np
from ase.utils import basestring
from ase.atoms import Atoms
from ase.calculators.calculator import Calculator
from clease import CorrFunction
from clease.settings import ClusterExpansionSetting
# from clease.calculator.duplication_count_tracker import DuplicationCountTracker
from clease_cxx import PyCEUpdater


class MovedIgnoredAtomError(Exception):
    """Raised when ignored atoms is moved."""
    pass


class Clease(Calculator):
    """Class for calculating energy using CLEASE.

    Parameters:

    setting: `ClusterExpansionSetting` object

    eci: dict
        Dictionary containing cluster names and their ECI values

    init_cf: `None` or dictionary (optional)
        If the correlation function of Atoms object is known, one can supply
        its correlation function values such that the initial assessment step
        is skipped. The dictionary should contain cluster names (same as the
        ones provided in `eci`) and their correlation function values.

    logfile: file object or str
        If *logfile* is a string, a file with that name will be opened.
        Use '-' for stdout.
    """

    name = 'CLEASE'
    implemented_properties = ['energy']

    def __init__(self, setting, eci=None, init_cf=None,
                 logfile=None):
        Calculator.__init__(self)

        if not isinstance(setting, ClusterExpansionSetting):
            msg = "setting must be CEBulk or CECrystal object."
            raise TypeError(msg)
        self.parameters["eci"] = eci
        self.setting = setting
        self.corrFunc = CorrFunction(setting)
        self.eci = eci
        # store cluster names
        self.cf_names = list(eci.keys())

        # calculate init_cf or convert init_cf to array
        if init_cf is None or isinstance(init_cf, dict):
            self.init_cf = init_cf
        else:
            msg = "'init_cf' must be a dictionary containing cluster names "
            msg += "and their correlation function values."
            raise TypeError(msg)

        if self.init_cf is not None and len(self.eci) != len(self.init_cf):
            msg = "length of provided ECIs and correlation functions do not "
            msg += "match"
            raise ValueError(msg)

        # logfile
        if isinstance(logfile, basestring):
            if logfile == '-':
                logfile = sys.stdout
            else:
                logfile = open(logfile, 'a')
        self.logfile = logfile

        self.energy = None
        # reference atoms for calculating the cf and energy for new atoms
        self.atoms = None
        self.symmetry_group = None
        self.is_backround_index = None

        # C++ updater initialized when atoms are set
        self.updater = None

    def _set_norm_factors(self):
        """Set normalization factor for each cluster.

        The normalization factor only kicks in when the cell is too small and
        thus, include self-interactions. This methods corrects the impact of
        self-interactions.
        """
        symm_group = np.zeros(len(self.setting.atoms), dtype=np.uint8)
        for num, group in enumerate(self.setting.index_by_trans_symm):
            symm_group[group] = num

        cluster_list = self.setting.cluster_list
        for cluster in cluster_list:
            fig_keys = list(set(cluster.get_all_figure_keys()))
            num_occ = {}
            for key in fig_keys:
                num_occ[key] = cluster_list.num_occ_figure(key,
                    cluster.name, symm_group, self.setting.trans_matrix)
            num_fig_occ = cluster.num_fig_occurences
            norm_factors = {}
            for key in fig_keys:
                tot_num = num_occ[key]
                num_unique = len(set(key.split("-")))
                norm_factors[key] = \
                    float(tot_num)/(num_unique*num_fig_occ[key])

            norm_factor_list = []
            for fig in cluster.indices:
                key = cluster.get_figure_key(fig)
                norm_factor_list.append(norm_factors[key])
            cluster.info['normalization_factor'] = norm_factor_list

    def set_atoms(self, atoms):
        self.atoms = atoms

        if self.init_cf is None:
            self.init_cf = \
                self.corrFunc.get_cf_by_names(self.atoms, self.cf_names)

        if len(self.setting.atoms) != len(atoms):
            msg = "Passed Atoms object and setting.atoms should have same "
            msg += "number of atoms."
            raise ValueError(msg)

        if not np.allclose(atoms.positions, self.setting.atoms.positions):
            msg = "Positions of all atoms in the passed Atoms object and "
            msg += "setting.atoms should be the same. "
            raise ValueError(msg)

        self.symmetry_group = np.zeros(len(atoms), dtype=int)
        for symm, indices in enumerate(self.setting.index_by_trans_symm):
            self.symmetry_group[indices] = symm
        self.is_backround_index = np.zeros(len(atoms), dtype=np.uint8)
        self.is_backround_index[self.setting.background_indices] = 1

        self._set_norm_factors()
        self.updater = PyCEUpdater(self.atoms, self.setting, self.init_cf,
                                   self.eci, self.setting.cluster_list)

    def get_energy_given_change(self, system_changes):
        """
        Calculate the energy when the change is known. No checking will be
        performed.

        Parameters:

        system_changes: list
            System changes. See doc-string of
            `clease.montecarlo.observers.MCObserver`
        """
        self.update_cf(system_changes=system_changes)
        self.energy = self.updater.get_energy()
        self.results['energy'] = self.energy
        return self.energy

    def calculate(self, atoms, properties, system_changes):
        """Calculate the energy of the passed Atoms object.

        If accept=True, the most recently used atoms object is used as a
        reference structure to calculate the energy of the passed atoms.
        Returns energy.

        Parameters:

        atoms: Atoms object
            ASE Atoms object

        system_changes: list
            System changes. See doc-string of
            `clease.montecarlo.observers.MCObserver`
        """
        Calculator.calculate(self, atoms)
        self.update_energy()
        self.energy = self.updater.get_energy()
        self.results['energy'] = self.energy
        return self.energy

    def clear_history(self):
        self.updater.clear_history()

    def restore(self):
        """Restore the Atoms object to its original configuration and energy.

        This method reverts the Atoms object to its oldest state stored in
        memory. The state is restored to either
        (1) an initial state when the calculator was attached, or
        (2) the state at which the `clear_history()` method was invoked last
            time.

        NOTE: The maximum capacity for the history buffer is 1000 steps
        """
        self.updater.undo_changes()
        self.energy = self.updater.get_energy()
        self.results['energy'] = self.energy

    def update_energy(self):
        """Update correlation function and get new energy."""
        self.update_cf()
        self.energy = self.updater.get_energy()

    @property
    def indices_of_changed_atoms(self):
        """Return the indices of atoms that have been changed."""
        changed = self.updater.get_changed_sites(self.atoms)
        for index in changed:
            if self.is_backround_index[index] and \
                    self.setting.ignore_background_atoms:
                msg = "Atom with index {} is a background atom.".format(index)
                raise MovedIgnoredAtomError(msg)

        return changed

    def get_cf(self):
        """Return the correlation functions as a dict"""
        return self.updater.get_cf()

    def update_cf(self, system_changes=None):
        """Update correlation function based on the reference value.

        Paramters:

        system_changes: list
            System changes. See doc-string of
            `clease.montecarlo.observers.MCObserver`
        """
        if system_changes is None:
            swapped_indices = self.indices_of_changed_atoms
            symbols = self.updater.get_symbols()
            system_changes = [(x, symbols[x], self.atoms[x].symbol)
                              for x in swapped_indices]
        for change in system_changes:
            self.updater.update_cf(change)

    @property
    def cf(self):
        temp_cf = self.updater.get_cf()
        return [temp_cf[x] for x in self.cf_names]

    def log(self):
        """Write energy to log file."""
        if self.logfile is None:
            return True
        self.logfile.write('{}\n'.format(self.energy))
        self.logfile.flush()

    def update_eci(self, eci):
        """Update the ECI values.

        Parameters:

        eci: dict
            Dictionary with new ECIs
        """
        self.eci = eci
        self.updater.set_eci(eci)

    def get_singlets(self):
        return self.updater.get_singlets()



from kivy.app import App


class ReconfigDB(object):
    """Reconfigure settings and correlation functions stored in DB."""
    # reconfig_in_progress = False
    settings = None
    app = None
    status = None

    def apply_settings(self):

        if not self.app.root.ids.sm.get_screen('Settings').apply_settings():
            msg = "Couldn't initialize settings based on the values specified "
            msg += "in the fields on Concentration\nand Settings panels. "
            msg += "Please ensure that all of the values are correct."
            self.status.text = msg
            return False

        self.status.text = "Reconfiguring..."
        return True

    def reconfig_settings(self):
        if self.apply_settings():
            try:
                self.app.root.settings.reconfigure_settings()
                msg = "Cluster data updated for all templates.\nPlease also "
                msg += "reconfigure DB entries if there are any structures "
                msg += "stored in DB."
                self.status.text = msg

            except Exception as exc:
                self.status.text = str(exc)

        self.app.root.reconfig_in_progress = False

    def reconfig_db(self):
        if self.apply_settings():
            try:
                from clease import CorrFunction
                CorrFunction(self.app.root.settings).reconfigure_db_entries()
                msg = "All DB entries are reconfigured."
                self.status.text = msg

            except Exception as exc:
                self.status.text = str(exc)

        self.app.root.reconfig_in_progress = False

    def reconfig_settings_db(self):
        if self.apply_settings():
            try:
                from clease.tools import reconfigure
                reconfigure(self.app.root.settings)
                msg = "Settings andall DB entries are reconfigured."
                self.status.text = msg

            except Exception as exc:
                self.status.text = str(exc)

        self.app.root.reconfig_in_progress = False



import numpy as np
from copy import deepcopy
from clease.tools import equivalent_deco, nested_array2list, list2str
from clease.cluster_fingerprint import ClusterFingerprint


class Cluster(object):
    def __init__(self, name='c0', size=0, diameter=0.0, fingerprint=[],
                 ref_indx=0, indices=[], equiv_sites=[], trans_symm_group=0):
        self.name = name
        self.size = size
        self.diameter = diameter
        self.fp = fingerprint
        self.ref_indx = ref_indx
        self.indices = indices
        self.equiv_sites = equiv_sites
        self.group = trans_symm_group
        self.info = {}

    def __lt__(self, other):
        """Comparison operator."""
        fp_lt = self.fp < other.fp
        if fp_lt:
            assert self.name < other.name
        else:
            assert self.name >= other.name
        return fp_lt

    def __eq__(self, other):
        """Equality operator."""
        fp_equal = self.fp == other.fp
        if fp_equal:
            assert self.name == other.name
        return fp_equal

    def __ne__(self, other):
        return not self.__eq__(other)

    def equiv_deco(self, deco):
        return equivalent_deco(deco, self.equiv_sites)

    def todict(self):
        """Return a dictionary representation."""
        return {'indices': self.indices,
                'size': self.size,
                'symm': self.group,
                'diameter': self.diameter,
                'name': self.name,
                'fingerprint': self.fp,
                'ref_indx': self.ref_indx,
                'equiv_sites': self.equiv_sites,
                'info': self.info}

    def from_dict(self, data):
        self.indices = data['indices']
        self.size = data['size']
        self.group = data['symm']
        self.diameter = data['diameter']
        self.name = data['name']
        self.fp = data['fingerprint']
        self.ref_indx = data['ref_indx']
        self.equiv_sites = data['equiv_sites']
        self.info = data['info']

    @staticmethod
    def load(data):
        cluster = Cluster(None, None, None, None, None, None, None, None)
        fp = ClusterFingerprint.load(data['fingerprint'])
        data['fingerprint'] = fp
        data['indices'] = nested_array2list(data['indices'])
        data['equiv_sites'] = nested_array2list(data['equiv_sites'])
        cluster.from_dict(data)
        return cluster

    def is_subcluster(self, other):
        """Check if the passed cluster is a subcluster of the current."""
        if len(self.indices) == 0:
            return True

        if len(self.indices[0]) >= len(other.indices[0]):
            return False

        return any(set(s1).issubset(s2) for s1 in self.indices
                   for s2 in other.indices)

    def get_figure(self, atoms):
        figure = atoms[self.indices[0]]
        figure_center = np.mean(figure.get_positions(), axis=0)
        cell_center = np.mean(figure.get_cell(), axis=0)/2.0
        figure.translate(cell_center - figure_center)
        figure.wrap()
        return figure

    def __str__(self):
        str_rep = 'Name: {}\n'.format(self.name)
        str_rep += 'Diameter: {}\n'.format(self.diameter)
        str_rep += 'Size: {}\n'.format(self.size)
        str_rep += 'Ref. indx: {}\n'.format(self.ref_indx)
        str_rep += 'Trans. symm group: {}\n'.format(self.group)
        str_rep += 'Indices: {}\n'.format(self.indices)
        str_rep += 'Equiv. sites: {}\n'.format(self.equiv_sites)
        str_rep += 'Fingerprint: {}\n'.format(self.fp)
        str_rep += 'Information: {}\n'.format(self.info)
        return str_rep

    def get_figure_key(self, figure):
        """Return a key representation of the figure."""
        return list2str(self._order_equiv_sites(figure))

    def _order_equiv_sites(self, figure):
        """Sort equivalent sites."""
        figure_cpy = deepcopy(figure)
        for eq_group in self.equiv_sites:
            equiv_indices = [figure[i] for i in eq_group]
            equiv_indices.sort()
            for count, i in enumerate(eq_group):
                figure_cpy[i] = equiv_indices[count]
        return figure_cpy

    @property
    def num_fig_occurences(self):
        """Number of currences for each figures."""
        occ_count = {}
        for figure in self.indices:
            key = self.get_figure_key(figure)
            current_num = occ_count.get(key, 0)
            occ_count[key] = current_num + 1
        return occ_count

    def corresponding_figure(self, ref_indx, target_figure, trans_matrix):
        """Find figures that correspond to another reference index.

        Parameters:

        ref_indx: int
            reference index

        taget_figres: list of indices
            list of atomic indices that constitute the original figure before
            translating

        tran_matrix: list of dicts
            translation matrix
        """
        for figure in self.indices:
            translated_figure = [trans_matrix[ref_indx][x] for x in figure]
            translated_figure = self._order_equiv_sites(translated_figure)
            if translated_figure == target_figure:
                return self._order_equiv_sites(figure)

        raise RuntimeError("There are no matching figure!")

    def get_all_figure_keys(self):
        return [self.get_figure_key(fig) for fig in self.indices]



from clease.montecarlo import Montecarlo
from clease.montecarlo.observers import SGCObserver
import numpy as np
from ase.units import kB


class InvalidChemicalPotentialError(Exception):
    pass


class SGCMonteCarlo(Montecarlo):
    """
    Class for running Monte Carlo in the Semi-Grand Canonical Ensebmle
    (i.e., fixed number of atoms, but varying composition)

    See the docstring of `clease.montecarlo.Montecarlo`

    Parameters:

    atoms: Atoms
        Atoms object (with CLEASE calculator attached!)

    temp: float
        Temperature in kelvin

    symbols: list
        Possible symbols to be used in swaps
    """

    def __init__(self, atoms, temp, symbols=[]):
        Montecarlo.__init__(self, atoms, temp)

        self.symbols = symbols

        if len(self.symbols) <= 1:
            raise ValueError("At least 2 symbols have to be specified")
        self.averager = SGCObserver(self.atoms.get_calculator())

        self.chem_pots = []
        self.chem_pot_names = []
        self.has_attached_avg = False
        self.name = "SGCMonteCarlo"
        self._chemical_potential = None
        self.chem_pot_in_eci = False
        self.current_singlets = None

        has_attached_obs = False
        for obs in self.observers:
            if obs.name == "SGCObserver":
                has_attached_obs = True
                self.averager = obs
                break
        if not has_attached_obs:
            self.attach(self.averager)

    def _get_trial_move(self):
        """Generate a trial move by flipping the symbol of one atom."""
        self.current_singlets = self.atoms.get_calculator().get_singlets()
        indx = np.random.randint(low=0, high=len(self.atoms))
        old_symb = self.atoms[indx].symbol
        new_symb = old_symb
        while new_symb == old_symb:
            new_symb = self.symbols[np.random.randint(
                low=0, high=len(self.symbols))]
        system_changes = [(indx, old_symb, new_symb)]
        return system_changes

    def _check_symbols(self):
        """
        Override because there are no restriction on the symbols here
        """
        pass

    def _update_tracker(self, system_changes):
        """
        Override the update of the atom tracker.

        The atom tracker is irrelevant in the semi grand canonical ensemble

        Parameter:

        system_changes: list
            Accepted system changes
        """
        pass

    def reset(self):
        """
        Reset the simulation object
        """
        super(SGCMonteCarlo, self).reset()
        self.averager.reset()

    @property
    def chemical_potential(self):
        return self._chemical_potential

    @chemical_potential.setter
    def chemical_potential(self, chem_pot):
        eci = self.atoms.get_calculator().eci
        if any([k not in eci.keys() for k in chem_pot.keys()]):
            msg = "A chemical potential not being trackted is added. Make "
            msg += "sure that all the following keys are in the ECIs before "
            msg += "they are passed to the calculator: "
            msg += "{}\n".format(list(chem_pot.keys()))
            msg += "(Add them with a zero ECI value if they are not supposed "
            msg += "to be included.)"
            raise InvalidChemicalPotentialError(msg)

        self._chemical_potential = chem_pot
        if self.chem_pot_in_eci:
            self._reset_eci_to_original(self.atoms.get_calculator().eci)
        self._include_chemical_potential_in_eci(
            chem_pot, self.atoms.get_calculator().eci)

    def _include_chemical_potential_in_eci(self, chem_pot, eci):
        """
        Including the chemical potentials in the ECIs

        Parameters:


        chem_pot: dict
            Chemical potentials
        eci: dict
            Original ECIs
        """
        self.chem_pots = []
        self.chem_pot_names = []
        keys = list(chem_pot.keys())
        keys.sort()
        for key in keys:
            self.chem_pots.append(chem_pot[key])
            self.chem_pot_names.append(key)
            current_eci = eci.get(key, 0.0)
            eci[key] = current_eci - chem_pot[key]
        calc = self.atoms.get_calculator()
        calc.update_eci(eci)
        self.chem_pot_in_eci = True
        self.current_energy = calc.calculate(None, None, None)
        return eci

    def _reset_eci_to_original(self, eci_with_chem_pot):
        """
        Resets the ECIs to their original value

        :parma dict eci_with_chem_pot: ECIs with chemical potential included
        """
        for name, val in zip(self.chem_pot_names, self.chem_pots):
            eci_with_chem_pot[name] += val
        calc = self.atoms.get_calculator()
        calc.update_eci(eci_with_chem_pot)
        self.chem_pot_in_eci = False
        self.current_energy = calc.calculate(None, None, None)
        return eci_with_chem_pot

    def reset_eci(self):
        """Return the ECIs."""
        if self.chem_pot_in_eci:
            self._reset_eci_to_original(self.atoms.get_calculator().eci)

    def run(self, steps=10, chem_pot=None):
        """
        Run Monte Carlo simulation.
        See :py:meth:`cemc.mcmc.Montecarlo.runMC`

        Parameters:

        chem_pot: dict
            Chemical potentials. The keys should correspond to one of the
            singlet terms. A typical form of this is {"c1_0":-1.0,c1_1_1.0}
        """

        if chem_pot is None and self.chemical_potential is None:
            ex_chem_pot = {
                "c1_1": -0.1,
                "c1_2": 0.05
            }
            raise ValueError("No chemicalpotentials given. Has to be "
                             "dictionary of the form {}".format(ex_chem_pot))

        if chem_pot is not None:
            self.chemical_potential = chem_pot
        self.reset()

        Montecarlo.run(self, steps=steps)

    def singlet2composition(self, avg_singlets):
        """Convert singlets to composition."""
        bf = self.atoms.get_calculator().setting.basis_functions
        matrix = np.zeros((len(self.symbols), len(self.symbols)))

        index = {s: i for i, s in enumerate(self.symbols)}
        for i, b in enumerate(bf):
            for s, col in index.items():
                matrix[i, col] = b[s]

        matrix[-1, :] = 1.0
        rhs = np.zeros(len(self.symbols))
        rhs[:-1] = avg_singlets
        rhs[-1] = 1.0
        x = np.linalg.solve(matrix, rhs)

        res = {}
        for s, i in index.items():
            name = s + "_conc"
            res[name] = x[i]
        return res

    def get_thermodynamic_quantities(self, reset_eci=True):
        """Compute thermodynamic quantities.

        Parameters:

        reset_eci: bool
            If True, the chemical potential will be removed from the ECIs
        """
        N = self.averager.counter
        quantities = {}
        singlets = self.averager.singlets/N
        singlets_sq = self.averager.quantities["singlets_sq"]/N

        quantities["sgc_energy"] = self.averager.energy.mean + self.energy_bias
        quantities["sgc_heat_capacity"] = self.averager.energy_sq.mean - \
            self.averager.energy.mean**2

        quantities["sgc_heat_capacity"] /= (kB*self.T**2)

        quantities["energy"] = self.averager.energy.mean + self.energy_bias
        natoms = len(self.atoms)
        for i in range(len(self.chem_pots)):
            quantities["energy"] += self.chem_pots[i]*singlets[i]*natoms

        quantities["temperature"] = self.T
        quantities["n_mc_steps"] = self.averager.counter

        # Add singlets and chemical potential to the dictionary
        for i in range(len(singlets)):
            name = "singlet_{}".format(self.chem_pot_names[i])
            quantities[name] = singlets[i]

            name = "var_singlet_{}".format(self.chem_pot_names[i])
            quantities[name] = singlets_sq[i] - singlets[i]**2

            name = "mu_{}".format(self.chem_pot_names[i])
            quantities[name] = self.chem_pots[i]

        quantities.update(self.meta_info)

        try:
            avg_conc = self.singlet2composition(singlets)
            quantities.update(avg_conc)
        except Exception as exc:
            print("Could not find average singlets!")
            print(exc)

        if reset_eci:
            self._reset_eci_to_original(self.atoms.get_calculator().eci)
        return quantities



# distutils: language = c++

from ce_updater cimport CEUpdater
from libcpp.vector cimport vector
from libcpp.string cimport string
from libcpp.map cimport map as map_cpp
from cython.operator import dereference, postincrement

cdef class PyCEUpdater:
    """Cython wrapper for the C++ class"""
    cdef CEUpdater *thisptr
    cdef object setting
    cdef object corr_func
    cdef object eci

    def __cinit__(self):
        self.thisptr = new CEUpdater()

    def __dealloc__(self):
        del self.thisptr

    def __init__(self, atoms, setting, corr_func, eci, cluster_info):
        self.setting = setting
        self.corr_func = corr_func
        self.eci = eci
        self.thisptr.init(atoms, setting, corr_func, eci, cluster_info)

    def clear_history(self):
        self.thisptr.clear_history()

    def undo_changes(self):
        self.thisptr.undo_changes()

    def update_cf(self, system_changes):
        self.thisptr.update_cf(system_changes)

    def calculate(self, system_changes):
        return self.thisptr.calculate(system_changes)

    def get_cf(self):
        return self.thisptr.get_cf()

    def set_eci(self, eci):
        self.thisptr.set_eci(eci)

    def get_singlets(self):
        return self.thisptr.get_singlets()

    def get_energy(self):
        return self.thisptr.get_energy()

    def get_symbols(self):
        return self.thisptr.get_symbols()

    def set_num_threads(self, num_threads):
        self.thisptr.set_num_threads(num_threads)

    def get_changed_sites(self, atoms):
        symbs = [atom.symbol for atom in atoms]
        cdef vector[string] symb_vec

        for atom in atoms:
            symb_vec.push_back(atom.symbol)

        cdef vector[unsigned int] changed
        self.thisptr.get_changes(symb_vec, changed)
        return [changed[i] for i in range(changed.size())]

    def calculate_cf_from_scratch(self, atoms, cf_names):
        self.thisptr.set_atoms(atoms)

        cdef map_cpp[string, double] cf
        cdef vector[string] cname_vec

        # Transfer names to a C++ vector
        for name in cf_names:
            cname_vec.push_back(name)
        self.thisptr.calculate_cf_from_scratch(cname_vec, cf)

        # Transfer to python dict
        cf_dict = {}
        cdef map_cpp[string, double].iterator it = cf.begin()
        cdef map_cpp[string, double].iterator end = cf.end()

        # Transfer map[string, double] to dictionary
        while(it != end):
            cf_dict[dereference(it).first] = dereference(it).second
            postincrement(it)
        return cf_dict



"""Definitions of Cluster Expansion settings for bulk.

Cluster Expansion can be peformed on bulk material using either CEBulk
or CECrystal class.
"""
from ase.build import bulk
from ase.spacegroup import crystal
from clease.tools import wrap_and_sort_by_position
from clease.settings import ClusterExpansionSetting
from copy import deepcopy


class CEBulk(ClusterExpansionSetting):
    """
    Specify cluster expansion settings for bulk materials defined based on
    crystal structures.

    Parameters:

    crystalstructure: str
        Must be one of sc, fcc, bcc, hcp, diamond, zincblende, rocksalt,
        cesiumchloride, fluorite or wurtzite.

    a: float
        Lattice constant.

    c: float
        Lattice constant.

    covera: float
        c/a ratio used for hcp.  Default is ideal ratio: sqrt(8/3).

    u: float
        Internal coordinate for Wurtzite structure.

    size: list
        Size of the supercell (e.g., [2, 2, 2] for 2x2x2 cell).

    supercell_factor: int
        Maximum multipilicity factor for limiting the size of supercell
        created from the primitive cell.

    concentration: Concentration object
        Concentration object specifying the concentration range of
        constituting species

    db_name: str
        Name of the database file

    max_cluster_size: int
        Maximum size (number of atoms in a cluster)

    max_cluster_dia: list of int or float
        A list of int or float containing the maximum diameter of clusters
        (in Å)

    basis_function: str
        One of "polynomial", "trigonometric" or "binary-linear"

    skew_threshold: int
        The maximum acceptable skew level (ratio of max and min diagonals of
        the cell) for supercells generated. A higher number allows highly
        skewed cells (e.g., 1x1x12) cells to be generated.

    ignore_background_atoms: bool
        if ``True``, a basis consisting of only one element type will be
        ignored when creating clusters.
    """

    def __init__(self, crystalstructure=None,
                 a=None, c=None, covera=None, u=None,
                 size=None, supercell_factor=None,
                 concentration=None, db_name=None, max_cluster_size=4,
                 max_cluster_dia=[5.0, 5.0, 5.0], basis_function='polynomial',
                 skew_threshold=4, ignore_background_atoms=False):

        # Initialization
        self.structures = {'sc': 1, 'fcc': 1, 'bcc': 1, 'hcp': 1, 'diamond': 1,
                           'zincblende': 2, 'rocksalt': 2, 'cesiumchloride': 2,
                           'fluorite': 3, 'wurtzite': 2}
        self.crystalstructure = crystalstructure
        self.a = a
        self.c = c
        self.covera = covera
        self.u = u

        ClusterExpansionSetting.__init__(self, size, supercell_factor,
                                         concentration, db_name,
                                         max_cluster_size, max_cluster_dia,
                                         basis_function, skew_threshold,
                                         ignore_background_atoms)

        # Save raw input arguments for save/load. The arguments gets altered
        # during the initalization process to handle 'ignore_background_atoms'
        # case
        self.kwargs.update({'crystalstructure': crystalstructure,
                            'a': a,
                            'c': c,
                            'covera': covera,
                            'u': u})
        num_basis = len(self.concentration.orig_basis_elements)
        if num_basis != self.structures[self.crystalstructure]:
            msg = "{} has {} basis. ".format(
                self.crystalstructure, self.structures[self.crystalstructure])
            msg += "The number of basis specified by basis_elements is "
            msg += "{}".format(num_basis)
            raise ValueError(msg)

        self._check_first_elements()

    def _get_prim_cell(self):
        basis_elements = self.concentration.orig_basis_elements
        num_basis = len(basis_elements)
        if num_basis == 1:
            atoms = bulk(name='{}'.format(basis_elements[0][0]),
                         crystalstructure=self.crystalstructure, a=self.a,
                         c=self.c, covera=self.covera, u=self.u)

        elif num_basis == 2:
            atoms = bulk(name='{}{}'.format(basis_elements[0][0],
                                            basis_elements[1][0]),
                         crystalstructure=self.crystalstructure, a=self.a,
                         c=self.c, covera=self.covera, u=self.u)

        else:
            atoms = bulk(name='{}{}{}'.format(basis_elements[0][0],
                                              basis_elements[1][0],
                                              basis_elements[2][0]),
                         crystalstructure=self.crystalstructure, a=self.a,
                         c=self.c, covera=self.covera, u=self.u)
        atoms = wrap_and_sort_by_position(atoms)
        return atoms

    @staticmethod
    def load(filename):
        """Load settings from a file in JSON format.

        Parameters:

        filename: str
            Name of the file that has the settings.
        """
        import json
        with open(filename, 'r') as infile:
            kwargs = json.load(infile)
        classtype = kwargs.pop("classtype")
        if classtype != 'CEBulk':
            raise TypeError('Loaded setting file is not for CEBulk class')
        return CEBulk(**kwargs)


class CECrystal(ClusterExpansionSetting):
    """Store CE settings on bulk materials defined based on space group.

    Parameters:

    basis: list of scaled coordinates
        Positions of the unique sites corresponding to symbols given
        either as scaled positions or through an atoms instance.

    spacegroup: int | string | Spacegroup instance
        Space group given either as its number in International Tables
        or as its Hermann-Mauguin symbol.

    cell: 3x3 matrix
        Unit cell vectors.

    cellpar: [a, b, c, alpha, beta, gamma]
        Cell parameters with angles in degree. Is not used when `cell`
        is given.

    ab_normal: vector
        Is used to define the orientation of the unit cell relative
        to the Cartesian system when `cell` is not given. It is the
        normal vector of the plane spanned by a and b.

    size: list
        Size of the supercell (e.g., [2, 2, 2] for 2x2x2 cell).

    supercell_factor: int
        Maximum multipilicity factor for limiting the size of supercell
        created from the primitive cell.

    concentration: Concentration object
        Concentration object specifying the concentration range of
        constituting species

    db_name: str
        name of the database file

    max_cluster_size: int
        maximum size (number of atoms in a cluster)

    max_cluster_dia: list of int or float
        A list of int or float containing the maximum diameter of clusters
        (in Å)

    basis_function: str
        One of "polynomial", "triogonometric" or "binary-linear"

    skew_threshold: int
        The maximum acceptable skew level (ratio of max and min diagonals of
        the cell) for supercells generated. A higher number allows highly
        skewed cells (e.g., 1x1x12) cells to be generated.

    ignore_background_atoms: bool
        if ``True``, a basis consisting of only one element type will be
        ignored when creating clusters.
    """

    def __init__(self, basis=None, spacegroup=1,
                 cell=None, cellpar=None, ab_normal=(0, 0, 1), size=None,
                 supercell_factor=None, concentration=None, db_name=None,
                 max_cluster_size=4, max_cluster_dia=[5.0, 5.0, 5.0],
                 basis_function='polynomial', skew_threshold=4,
                 ignore_background_atoms=False):

        # Initialization
        self.basis = basis
        self.spacegroup = spacegroup
        self.cell = cell
        self.cellpar = cellpar
        self.ab_normal = ab_normal
        self.symbols = []
        num_basis = len(concentration.orig_basis_elements)
        for x in range(num_basis):
            self.symbols.append(concentration.orig_basis_elements[x][0])

        ClusterExpansionSetting.__init__(self, size, supercell_factor,
                                         concentration, db_name,
                                         max_cluster_size, max_cluster_dia,
                                         basis_function, skew_threshold,
                                         ignore_background_atoms)

        # Save raw input arguments for save/load. The arguments gets altered
        # during the initalization process to handle 'ignore_background_atoms'
        # case
        self.kwargs.update({'basis': deepcopy(basis),
                            'spacegroup': spacegroup,
                            'cell': cell,
                            'cellpar': cellpar,
                            'ab_normal': ab_normal})

        self._check_first_elements()

    def _get_prim_cell(self):
        atoms = crystal(symbols=self.symbols, basis=self.basis,
                        spacegroup=self.spacegroup, cell=self.cell,
                        cellpar=self.cellpar, ab_normal=self.ab_normal,
                        size=[1, 1, 1], primitive_cell=True)
        atoms = wrap_and_sort_by_position(atoms)
        return atoms

    @staticmethod
    def load(filename):
        """Load settings from a file in JSON format.

        Parameters:

        filename: str
            Name of the file that has the settings.
        """
        import json
        with open(filename, 'r') as infile:
            kwargs = json.load(infile)
        classtype = kwargs.pop("classtype")
        if classtype != 'CECrystal':
            raise TypeError('Loaded setting file is not for CEBulk class')
        return CEBulk(**kwargs)



from kivy.uix.screenmanager import Screen
from kivy.uix.popup import Popup
from kivy.utils import get_color_from_hex
from kivy.app import App

from clease.gui.constants import INACTIVE_TEXT_COLOR, FOREGROUND_TEXT_COLOR
from clease.gui.load_save_dialog import LoadDialog
from clease.gui.util import parse_max_cluster_dia, parse_size
from clease.gui.util import parse_cell, parse_coordinate_basis, parse_cellpar
from clease.gui.settingInitializer import SettingsInitializer
from threading import Thread


class SettingsPage(Screen):
    _pop_up = None

    def dismiss_popup(self):
        if self._pop_up is None:
            return
        self._pop_up.dismiss()
        self._pop_up = None

    def show_load_dialog(self):
        content = LoadDialog(load=self.load, cancel=self.dismiss_popup)
        self._pop_up = Popup(title="Load structure DB", content=content,
                             pos_hint={'right': 0.95, 'top': 0.95},
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def to_dict(self):
        return {'type': self.ids.typeSpinner.text,
                'basis_function': self.ids.bfSpinner.text,
                'cluster_size': self.ids.clusterSize.text,
                'max_cluster_dia': self.ids.clusterDia.text,
                'db_name': self.ids.dbNameInput.text,
                'crystalstructure': self.ids.crystStructSpinner.text,
                'aParameter': self.ids.aParameterInput.text,
                'cParameter': self.ids.cParameterInput.text,
                'uParameter': self.ids.uParameterInput.text,
                'cellpar': self.ids.cellParInput.text,
                'cell': self.ids.cellInput.text,
                'basis': self.ids.crdBasisInput.text,
                'spacegroup': self.ids.spInput.text,
                'cell_size': self.ids.sizeInput.text,
                'cell_mode_spinner': self.ids.sizeSpinner.text,
                'supercell_factor': self.ids.scFactorInput.text,
                'skewness_factor': self.ids.skewFactorInput.text}

    def from_dict(self, data):
        self.ids.typeSpinner.text = data['type']
        self.ids.bfSpinner.text = data['basis_function']
        self.ids.clusterSize.text = data['cluster_size']
        self.ids.clusterDia.text = data['max_cluster_dia']
        self.db_name = data['db_name']
        self.ids.dbNameInput.text = data['db_name']
        self.ids.crystStructSpinner.text = data['crystalstructure']
        self.ids.aParameterInput.text = data['aParameter']
        self.ids.cParameterInput.text = data['cParameter']
        self.ids.uParameterInput.text = data['uParameter']
        self.ids.cellParInput.text = data['cellpar']
        self.ids.cellInput.text = data['cell']
        self.ids.crdBasisInput.text = data['basis']
        self.ids.spInput.text = data['spacegroup']
        self.ids.sizeInput.text = data.get('cell_size', '3, 3, 3')
        self.ids.sizeSpinner.text = data.get('cell_mode_spinner', 'Fixed')
        self.ids.scFactorInput.text = data.get('supercell_factor', '20')
        self.ids.skewFactorInput.text = data.get('skewness_factor', '4')

    def load(self, path, filename):
        self.db_path = path

        if len(filename) == 0:
            self.ids.dbNameInput.text = path
        else:
            self.ids.dbNameInput.text = filename[0]
        self.dismiss_popup()

    def update_input_section(self, text):
        if text == 'CECrystal':
            self._disable_CEBulk()
            self._enable_CECrystal()
        else:
            self._enable_CEBulk()
            self._disable_CECrystal()

    def on_enter(self):
        self.update_input_section(self.ids.typeSpinner.text)
        self.update_size_section(self.ids.sizeSpinner.text)

    def _disable_CEBulk(self):
        color = get_color_from_hex(INACTIVE_TEXT_COLOR)
        self.ids.CEBulkTitle.color = color
        self.ids.crystStruct.color = color
        self.ids.aParameter.color = color
        self.ids.cParameter.color = color
        self.ids.uParameter.color = color

        # Disable
        self.ids.crystStructSpinner.disabled = True
        self.ids.aParameterInput.disabled = True
        self.ids.cParameterInput.disabled = True
        self.ids.uParameterInput.disabled = True

    def _disable_CECrystal(self):
        color = get_color_from_hex(INACTIVE_TEXT_COLOR)
        self.ids.CECrystalTitle.color = color
        self.ids.cell.color = color
        self.ids.cellPar.color = color
        self.ids.spLabel.color = color
        self.ids.crBasisLabel.color = color

        # Disable fields
        self.ids.cellInput.disabled = True
        self.ids.cellParInput.disabled = True
        self.ids.spInput.disabled = True
        self.ids.crdBasisInput.disabled = True

    def _enable_CEBulk(self):
        color = get_color_from_hex(FOREGROUND_TEXT_COLOR)
        self.ids.CEBulkTitle.color = color
        self.ids.crystStruct.color = color
        self.ids.aParameter.color = color
        self.ids.cParameter.color = color
        self.ids.uParameter.color = color

        # Enable
        self.ids.crystStructSpinner.disabled = False
        self.ids.aParameterInput.disabled = False
        self.ids.cParameterInput.disabled = False
        self.ids.uParameterInput.disabled = False

    def _enable_CECrystal(self):
        color = get_color_from_hex(FOREGROUND_TEXT_COLOR)
        self.ids.CECrystalTitle.color = color
        self.ids.cell.color = color
        self.ids.cellPar.color = color
        self.ids.spLabel.color = color
        self.ids.crBasisLabel.color = color

        # Disable fields
        self.ids.cellInput.disabled = False
        self.ids.cellParInput.disabled = False
        self.ids.spInput.disabled = False
        self.ids.crdBasisInput.disabled = False

    def _check_cebulk_parameters(self):
        """Check the user input of the CE bulk parameters."""
        try:
            _ = float(self.ids.aParameterInput.text)
        except Exception:
            App.get_running_app().root.ids.status.text =\
                "a has to be a float"
            return 1

        c = self.ids.cParameterInput.text
        if c != '':
            try:
                _ = float(c)
            except Exception:
                App.get_running_app().root.ids.status.text =\
                    "c has to be float"
                return 1

        u = self.ids.uParameterInput.text

        if u != '':
            try:
                _ = float(u)
            except Exception:
                App.get_running_app().root.ids.status.text =\
                    "u has to be float"
                return 1
        return 0

    def cellpar_ok(self):
        cellPar = self.ids.cellParInput.text
        try:
            _ = parse_cellpar(cellPar)
        except Exception as exc:
            App.get_running_app().root.ids.status.text = str(exc)
            return False
        return True

    def cell_ok(self):
        cell = self.ids.cellInput.text
        try:
            _ = parse_cell(cell)
        except Exception as exc:
            App.get_running_app().root.ids.status.text = str(exc)
            return False
        return True

    def coordinate_basis_ok(self):
        crdBasis = self.ids.crdBasisInput.text
        try:
            _ = parse_coordinate_basis(crdBasis)
        except Exception as exc:
            App.get_running_app().root.ids.status.text = str(exc)
            return False
        return True

    def _check_cecrystal_input(self):
        cellPar = self.ids.cellParInput.text
        sufficient_cell_info_given = False
        if cellPar != '':
            if not self.cellpar_ok():
                return 1
            sufficient_cell_info_given = True

        cell = self.ids.cellInput.text
        if cell != '':
            if not self.cell_ok():
                return 1
            sufficient_cell_info_given = True

        if not sufficient_cell_info_given:
            msg = 'Either cellpar or cell has to be given'
            App.get_running_app().root.ids.status.text = msg
            return 1

        if not self.coordinate_basis_ok():
            return 1

        try:
            _ = int(self.ids.spInput.text)
        except Exception:
            msg = "Spacegroup has to be an integer."
            App.get_running_app().root.ids.status.text = msg
            return 1
        return 0

    def max_cluster_dia_ok(self):
        cluster_size = int(self.ids.clusterSize.text)
        try:
            diameter = parse_max_cluster_dia(self.ids.clusterDia.text)
            if isinstance(diameter, list):
                if len(diameter) != cluster_size - 1:
                    msg = 'Cluster dia has to be given for 2-body and beyond!'
                    App.get_running_app().root.ids.status.text = msg
                    return False
        except Exception as exc:
            App.get_running_app().root.ids.status.text = str(exc)
            return False
        return True

    def cell_size_ok(self):
        """Check if the cell size is OK."""
        try:
            _ = parse_size(self.ids.sizeInput.text)
        except Exception as exc:
            App.get_running_app().root.ids.status.text = str(exc)
            return False
        return True

    def check_user_input(self):
        """Check the input values from the user."""

        # Check max cluster size
        try:
            _ = int(self.ids.clusterSize.text)
        except Exception:
            msg = "Max cluster size has to be an integer."
            App.get_running_app().root.ids.status.text = msg
            return 1

        # Check that we can parse the max cluster diameter
        if not self.max_cluster_dia_ok():
            return 1

        # Check that we can parse size
        if self.ids.sizeSpinner.text == 'Fixed':
            if not self.cell_size_ok():
                return 1
        else:
            if self.ids.scFactorInput.text == '':
                msg = 'Supercell factor has to be given'
                App.get_running_app().root.ids.status.text = msg
                return 1

            if self.ids.skewFactorInput.text == '':
                msg = 'Skewness factor has to be given'
                App.get_running_app().root.ids.status.text = msg
                return 1

        db_name = self.ids.dbNameInput.text
        if db_name == '':
            msg = "No database given"
            App.get_running_app().root.ids.status.text = msg
            return 1

        if self.ids.typeSpinner.text == "CEBulk":
            error_code = self._check_cebulk_parameters()

            if error_code != 0:
                return error_code
        else:
            error_code = self._check_cecrystal_input()
            if error_code != 0:
                return error_code

        return 0

    def update_size_section(self, text):
        inactive = get_color_from_hex(INACTIVE_TEXT_COLOR)
        active = get_color_from_hex(FOREGROUND_TEXT_COLOR)
        if text == 'Fixed':
            self.ids.skewnessLabel.color = inactive
            self.ids.supercellFactLabel.color = inactive
            self.ids.sizeLabel.color = active

            self.ids.scFactorInput.disabled = True
            self.ids.skewFactorInput.disabled = True
            self.ids.sizeInput.disabled = False
        else:
            self.ids.skewnessLabel.color = active
            self.ids.supercellFactLabel.color = active
            self.ids.sizeLabel.color = inactive

            self.ids.scFactorInput.disabled = False
            self.ids.skewFactorInput.disabled = False
            self.ids.sizeInput.disabled = True

    def apply_settings(self):
        try:
            from clease import Concentration
            conc_page = self.manager.get_screen("Concentration")
            if conc_page.check_user_input() != 0:
                return

            A_lb, rhs_lb, A_eq, rhs_eq = conc_page.get_constraint_matrices()

            if not conc_page.elements:
                msg = 'It appears that the Apply button in Concentration panel '
                msg += 'was not clicked.'
                App.get_running_app().root.ids.status.text = msg
                return
            basis_elements = conc_page.elements
            grouped_basis = conc_page.grouped_basis
            conc = Concentration(basis_elements=basis_elements,
                                 A_lb=A_lb, b_lb=rhs_lb,
                                 A_eq=A_eq, b_eq=rhs_eq,
                                 grouped_basis=grouped_basis)

            if self.check_user_input() != 0:
                return

            settingPage = self.to_dict()
            supercell_factor = int(settingPage['supercell_factor'])
            skewness_factor = int(settingPage['skewness_factor'])
            size = None

            if self.ids.sizeSpinner.text == 'Fixed':
                size = parse_size(settingPage['cell_size'])

            if settingPage['max_cluster_dia'] == '':
                max_cluster_dia = None
            else:
                max_cluster_dia =\
                    parse_max_cluster_dia(settingPage['max_cluster_dia'])

            initializer = SettingsInitializer()
            initializer.app = App.get_running_app()
            initializer.status = App.get_running_app().root.ids.status

            if settingPage["type"] == 'CEBulk':
                if settingPage['aParameter'] == '':
                    a = None
                else:
                    a = float(settingPage['aParameter'])

                if settingPage['cParameter'] == '':
                    c = None
                else:
                    c = float(settingPage['cParameter'])

                if settingPage['uParameter'] == '':
                    u = None
                else:
                    u = float(settingPage['uParameter'])
                kwargs = dict(
                    crystalstructure=settingPage['crystalstructure'], a=a,
                    c=c, u=u,
                    db_name=settingPage['db_name'], concentration=conc,
                    max_cluster_dia=max_cluster_dia,
                    max_cluster_size=int(settingPage['cluster_size']),
                    basis_function=settingPage['basis_function'],
                    size=size, supercell_factor=supercell_factor,
                    skew_threshold=skewness_factor
                )
                msg = "Applying settings to database..."
                App.get_running_app().root.ids.status.text = msg
                initializer.type = 'CEBulk'
                initializer.kwargs = kwargs
                Thread(target=initializer.initialize).start()
            else:
                if settingPage['cellpar'] == '':
                    cellpar = None
                else:
                    cellpar = parse_cellpar(settingPage['cellpar'])

                if settingPage['basis'] == '':
                    basis = None
                else:
                    basis = parse_coordinate_basis(settingPage['basis'])

                if settingPage['cell'] == '':
                    cell = None
                else:
                    cell = parse_cell(settingPage['cell'])


                sp = int(settingPage['spacegroup'])
                msg = "Applying settings to database..."
                App.get_running_app().root.ids.status.text = msg
                kwargs = dict(
                    basis=basis, cellpar=cellpar, cell=cell,
                    max_cluster_dia=max_cluster_dia,
                    max_cluster_size=int(settingPage['cluster_size']),
                    basis_function=settingPage['basis_function'],
                    size=size, supercell_factor=supercell_factor,
                    skew_threshold=skewness_factor,
                    concentration=conc, db_name=settingPage['db_name'],
                    spacegroup=sp
                )
                initializer.type = 'CECrystal'
                initializer.kwargs = kwargs
                Thread(target=initializer.initialize).start()
                msg = "Settings initialized."
                App.get_running_app().root.ids.status.text = msg
            return True

        except Exception as exc:
            App.get_running_app().root.ids.status.text = str(exc)
            return False



from kivy.uix.floatlayout import FloatLayout
from kivy.properties import ObjectProperty
from kivy.uix.textinput import TextInput
from kivy.uix.spinner import Spinner
import os


class FitAlgEditor(FloatLayout):
    close = ObjectProperty(None)
    backup_file = 'fit_alg_editor.txt'
    backup_folder = '.cleaseGUI/'

    def __init__(self, **kwargs):
        FloatLayout.__init__(self, **kwargs)
        if not os.path.exists(self.backup_folder):
            os.mkdir(self.backup_folder)
        self.load_values()

    @property
    def backup_filename(self):
        return self.backup_folder + self.backup_file

    def backup(self):
        values = []
        for child in self.ids.mainLayout.children:
            if isinstance(child, (TextInput, Spinner)):
                txt = child.text
                values.append(txt)

        with open(self.backup_filename, 'w') as out:
            for v in values:
                out.write('{}\n'.format(v))

    def load_values(self):
        try:
            with open(self.backup_filename, 'r') as infile:
                lines = infile.readlines()
            lines = [line.strip() for line in lines]

            counter = 0
            for child in self.ids.mainLayout.children:
                if isinstance(child, (TextInput, Spinner)):
                    child.text = lines[counter]
                    counter += 1
        except Exception:
            pass


class LassoEditor(FitAlgEditor):
    backup_file = 'lasso_editor.txt'


class L2Editor(FitAlgEditor):
    backup_file = 'l2_editor.txt'


class BCSEditor(FitAlgEditor):
    backup_file = 'bcs_editor.txt'


class GAEditor(FitAlgEditor):
    backup_file = 'ga_editor.txt'



FOREGROUND_TEXT_COLOR = '#FFFFFF'
INACTIVE_TEXT_COLOR = '#333333'
BACKGROUND_COLOR = '#000000'
ECI_GRAPH_COLORS = ['#e41a1c', '#377eb8', '#4daf4a', '#984ea3', '#ff7f00', 
                    '#ffff33']
MC_MEAN_CURVE_COLOR = '#8da0cb'

# Modes for concentration in MC
SYSTEMS_FROM_DB = 0
CONC_PER_BASIS = 1


"""Module for setting up pseudospins and basis functions."""
import numpy as np
import math


class BasisFunction(object):
    """Base class for all Basis Functions."""

    def __init__(self, unique_elements):
        self.name = "generic"
        self.unique_elements = sorted(unique_elements)
        self.num_unique_elements = len(unique_elements)
        if self.num_unique_elements < 2:
            raise ValueError("Systems must have more than 1 type of element.")
        self.spin_dict = self.get_spin_dict()
        self.basis_functions = self.get_basis_functions()

    def get_spin_dict(self):
        """Get spin dictionary."""
        raise NotImplementedError("get_spin_dict has to be implemented in "
                                  "derived classes!")

    def get_basis_functions(self):
        """Get basis function."""
        raise NotImplementedError("get_basis_functions has to be implemented "
                                  "in derived classes!")

    def customize_full_cluster_name(self, full_cluster_name):
        """Customize the full cluster names. Default is to do nothing."""
        return full_cluster_name


class Polynomial(BasisFunction):
    """Pseudospin and basis function from Sanchez et al.

    Sanchez, J. M., Ducastelle, F. and Gratias, D. (1984).
    Generalized cluster description of multicomponent systems.
    Physica A: Statistical Mechanics and Its Applications, 128(1-2), 334-350.
    """

    def __init__(self, unique_elements):
        BasisFunction.__init__(self, unique_elements)
        self.name = "polynomial"
        if self.num_unique_elements > 6:
            raise ValueError("Only systems consisting up to 6 types of "
                             "elements are currently supported for the "
                             "Polynomial basis function scheme.")

    def get_spin_dict(self):
        """Define pseudospins for all consistuting elements."""
        # Find odd/even
        spin_values = []
        if self.num_unique_elements % 2 == 1:
            highest = (self.num_unique_elements - 1) / 2
        else:
            highest = self.num_unique_elements / 2
        # Assign spin value for each element
        while highest > 0:
            spin_values.append(highest)
            spin_values.append(-highest)
            highest -= 1
        if self.num_unique_elements % 2 == 1:
            spin_values.append(0)

        spin_dict = {}
        for x in range(self.num_unique_elements):
            spin_dict[self.unique_elements[x]] = spin_values[x]
        return spin_dict

    def get_basis_functions(self):
        """Create basis functions to guarantee the orthonormality."""
        # coeff_c = [c_0^1, c_1^1, c_0^2, c_1^2, c_2^2]^T
        coeff_c = np.array([
            [0.0, np.sqrt(2), -5. / 3, -1 * np.sqrt(10. / 7), -np.sqrt(2)],
            [0.0,  -3 / np.sqrt(2), 2. / 3, np.sqrt(5. / 14),
             3. / (7 * np.sqrt(2))],
            [0.0, 0.0, 0.0, 3 * np.sqrt(2. / 7), 9 * np.sqrt(1.5) / 5],
            [0.0, 0.0, 0.0, -155 / (12 * np.sqrt(14)),
             -101 / (28 * np.sqrt(6))],
            [0.0, 0.0, 0.0, 5 * np.sqrt(7. / 2) / 12, 7 / (20 * np.sqrt(6))]])

        # coeff_d = [d_0^0, d_0^1, d_1^1, d_0^2, d_1^2, d_2^2]^T
        coeff_d = np.array([
            [1.0, np.sqrt(3. / 2), np.sqrt(2. / 5), 1. / np.sqrt(2),
             -np.sqrt(3. / 14)],
            [0.0, 0.0,  -17 / (3 * np.sqrt(10)), -17 / (6 * np.sqrt(2)),
             -7. / 6],
            [0.0, 0.0, np.sqrt(2.5) / 3, 5 / (6 * np.sqrt(2)), 1. / 6],
            [0.0, 0.0, 0.0, 0.0, 131. / (15 * np.sqrt(14))],
            [0.0, 0.0, 0.0, 0.0, -7 * np.sqrt(7. / 2) / 12],
            [0.0, 0.0, 0.0, 0.0, np.sqrt(7. / 2) / 20]])

        col = self.num_unique_elements - 2
        if col > 5:
            msg = "Polynomial basis function scheme only supports up "
            msg += "to 6 elements. \nPlease change to another basis function "
            msg += "scheme."
            raise RuntimeError(msg)

        bf_list = []
        bf = {}
        for key, value in self.spin_dict.items():
            bf[key] = coeff_d[0][col] * value
        bf_list.append(bf)

        if self.num_unique_elements > 2:
            bf = {}
            for key, value in self.spin_dict.items():
                bf[key] = (coeff_c[0][col] + (coeff_c[1][col] * value**2))
            bf_list.append(bf)

        if self.num_unique_elements > 3:
            bf = {}
            for key, value in self.spin_dict.items():
                bf[key] = coeff_d[1][col] * value
                bf[key] += coeff_d[2][col] * (value**3)
            bf_list.append(bf)

        if self.num_unique_elements > 4:
            bf = {}
            for key, value in self.spin_dict.items():
                bf[key] = coeff_c[2][col] + (coeff_c[3][col] * (value**2))
                bf[key] += coeff_c[4][col] * (value**4)
            bf_list.append(bf)

        if self.num_unique_elements > 5:
            bf = {}
            for key, value in self.spin_dict.items():
                bf[key] = coeff_d[3][col] * value
                bf[key] += coeff_d[4][col] * (value**3)
                bf[key] += coeff_d[5][col] * (value**5)
            bf_list.append(bf)

        return bf_list


class Trigonometric(BasisFunction):
    """Pseudospin and basis function from van de Walle.

    van de Walle, A. (2009).
    Multicomponent multisublattice alloys, nonconfigurational entropy and other
    additions to the Alloy Theoretic Automated Toolkit. Calphad, 33(2),
    266-278.
    """

    def __init__(self, unique_elements):
        BasisFunction.__init__(self, unique_elements)
        self.name = "trigonometric"

    def get_spin_dict(self):
        """Define pseudospins for all consistuting elements."""
        spin_values = list(range(self.num_unique_elements))
        spin_dict = {}
        for x in range(self.num_unique_elements):
            spin_dict[self.unique_elements[x]] = spin_values[x]
        return spin_dict

    def get_basis_functions(self):
        """Create basis functions to guarantee the orthonormality."""
        alpha = list(range(1, self.num_unique_elements))
        bf_list = []

        for a in alpha:
            bf = {}
            for key, value in self.spin_dict.items():
                var = 2 * np.pi * math.ceil(a/2.) * value
                var /= self.num_unique_elements
                if a % 2 == 1:
                    bf[key] = -np.cos(var) + 0.
                else:
                    bf[key] = -np.sin(var) + 0.

            # normalize the basis function
            sum = 0
            for key, value in self.spin_dict.items():
                sum += bf[key] * bf[key]
            normalization_factor = np.sqrt(self.num_unique_elements / sum)

            for key, value in bf.items():
                bf[key] = value * normalization_factor

            bf_list.append(bf)

        return bf_list


def kronecker(i, j):
    """Kronecker delta function."""
    if i == j:
        return 1
    return 0


class BinaryLinear(BasisFunction):
    """Pseudospin and basis function from Zhang and Sluiter.

    Zhang, X. and Sluiter M.
    Cluster expansions for thermodynamics and kinetics of multicomponent
    alloys.
    Journal of Phase Equilibria and Diffusion 37(1) 44-52.
    """

    def __init__(self, unique_elements, reduntant_element="auto"):
        if reduntant_element == "auto":
            self.redundant_element = sorted(unique_elements)[0]
        else:
            self.redundant_element = reduntant_element
        BasisFunction.__init__(self, unique_elements)
        self.name = "binary-linear"

    def get_spin_dict(self):
        """Define pseudospins for all consistuting elements."""
        spin_values = list(range(self.num_unique_elements))
        spin_dict = {}
        for x in range(self.num_unique_elements):
            spin_dict[self.unique_elements[x]] = spin_values[x]
        return spin_dict

    def get_basis_functions(self):
        """Create orthonormal basis functions.

        Due to the constraint that any site is occupied by exactly one element,
        we only need to track N-1 species if there are N species.
        Hence, the first element specified is redundant, and will not
        have a basis function.
        """
        bf_list = []
        num_bf = self.num_unique_elements
        for bf_num in range(num_bf):
            if self.unique_elements[bf_num] == self.redundant_element:
                continue
            new_bf = {symb: float(kronecker(i, bf_num))
                      for i, symb in enumerate(self.unique_elements)}
            bf_list.append(new_bf)
        return bf_list

    def _decoration2element(self, dec_num):
        """Get the element with its basis function equal to 1."""
        bf = self.bf_list[dec_num]
        for k, v in bf.items():
            if v == 1:
                return k
        raise ValueError("Did not find any element where the value is 1.")

    def customize_full_cluster_name(self, full_cluster_name):
        """Translate the decoration number to element names."""
        dec = full_cluster_name.rpartition("_")[1]
        name = full_cluster_name.rpartition("_")[0]
        new_dec = ""
        for decnum in dec:
            element = self._decoration2element(int(decnum))
            new_dec += "{}".format(element)
        return name + "_" + new_dec



"""Class containing a manager for setting up concentrations of species."""
import numpy as np
from scipy.optimize import minimize
from collections import OrderedDict


class IntConversionNotConsistentError(Exception):
    """
    Exception raised if equality constraints are not satisfied when
    converted to integers.
    """
    pass


class InvalidConcentrationError(Exception):
    """Exception raised when no valid concentration could be found."""
    pass


class InvalidConstraintError(Exception):
    """Exception raised when user provides invalid constraints."""
    pass


class Concentration(object):
    """
    Specify concentration ranges of consituting elements for cluster
    expansion.
    Concentration range can be specified in three different ways.

        1. specifying the equality and lower bound conditions by specifying
        ``A_lb``, ``b_lb``, ``A_eq`` and ``b_eq`` during initialization,
        2. using ``set_conc_formula_unit()`` method, and
        3. using ``set_conc_ranges()`` method.


    Parameters:

    basis_elements: list
        List of chemical symbols of elements to occupy each basis.
        Even for the cases where there is only one basis (e.g., fcc, bcc, sc),
        a list of symbols should be grouped by basis as in [['Cu', 'Au']]
        (note the nested list form).

    grouped_basis: list (optional, only used when basis are grouped)
        Indices of basis_elements that are considered to be equivalent when
        specifying concentration (e.g., useful when two basis are shared by
        the same set of elements and no distinctions are made between them)

    A_lb: list (optional, only used for linear algebra representation)
        A two-dimention matrix (or nested list) used to specify the lower
        bounds of the concentration ranges.

    b_lb: list (optional, only used for linear algebra representation)
        A list used tp specify the lower bounds of the concentration ranges.

    A_eq: list (optional, only used for linear algegra representation)
        A two-dimention matrix (or nested list) used to specify the equality
        conditions of the concentration ranges.

    b_eq: list (optional, only used for linear algegra representation)
        A list used tp specify the equality condisitons of the concentration
        ranges.
    """

    def __init__(self, basis_elements=None, grouped_basis=None,
                 A_lb=None, b_lb=None, A_eq=None, b_eq=None):
        self.orig_basis_elements = basis_elements
        self.grouped_basis = grouped_basis
        self._check_grouped_basis_elements()
        self.basis_elements = self._get_grouped_basis_elements()

        num_implicit_eq = len(self.basis_elements)
        self.num_concs = len([x for sub in self.basis_elements for x in sub])
        self.fixed_element_constraint_added = False

        num_usr_lb = 0
        if A_lb is not None:
            A_lb = np.array(A_lb)
            num_usr_lb = A_lb.shape[0]
        self.A_lb = np.zeros((self.num_concs, self.num_concs), dtype=int)
        self.b_lb = np.zeros(self.num_concs, dtype=int)

        num_usr_eq = 0
        if A_eq is not None:
            A_eq = np.array(A_eq)
            num_usr_eq = A_eq.shape[0]

        self.A_eq = np.zeros((num_implicit_eq, self.num_concs), dtype=int)
        self.b_eq = np.zeros(num_implicit_eq, dtype=int)
        self.orig_num_equality = self.A_eq.shape[0]

        # Ensure concentration in each basis sum to 1
        start_col = 0
        for i, basis in enumerate(self.basis_elements):
            if start_col + len(basis) >= self.A_eq.shape[1]:
                self.A_eq[i, start_col:] = 1
            else:
                self.A_eq[i, start_col:start_col+len(basis)] = 1
            start_col += len(basis)
            self.b_eq[i] = 1

        # Ensure that we have only positive concentrations
        for i in range(self.num_concs):
            self.A_lb[i, i] = 1

        self._linked_basis = [i for i in range(len(self.basis_elements))]
        if num_usr_eq > 0:
            self.add_usr_defined_eq_constraints(A_eq, b_eq)

        if num_usr_lb > 0:
            self.add_usr_defined_ineq_constraints(A_lb, b_lb)

    def _remove_redundant_entries(self, A, b):
        for row_num in range(A.shape[0]-1, -1, -1):
            for prev_row in range(row_num-1):
                cond1 = np.allclose(A[row_num, :], A[prev_row, :])
                cond2 = abs(b[row_num] - b[prev_row]) < 1E-9
                if cond1 and cond2:
                    A = np.delete(A, row_num, axis=0)
                    b = np.delete(b, row_num)
                    break
        return A, b

    def _get_grouped_basis_elements(self):
        if self.grouped_basis is None:
            return self.orig_basis_elements

        gr_basis_elements = []
        for group in self.grouped_basis:
            gr_basis_elements.append(self.orig_basis_elements[group[0]])
        return gr_basis_elements

    def to_dict(self):
        """Return neescary information to store as JSON."""
        data = {
            "A_eq": self.A_eq.tolist(),
            "b_eq": self.b_eq.tolist(),
            "A_lb": self.A_lb.tolist(),
            "b_lb": self.b_lb.tolist(),
            "basis_elements": self.orig_basis_elements,
        }
        if self.grouped_basis is not None:
            data["grouped_basis"] = self.grouped_basis
        return data

    @staticmethod
    def from_dict(data):
        """Initialize data from dictionary."""
        return Concentration(**data)

    def is_valid_conc(self, conc):
        """Check if the concentration is valid."""
        eq_valid = np.allclose(self.A_eq.dot(conc), self.b_eq)
        ineq_valid = np.all(self.A_lb.dot(conc) - self.b_lb > -1e-6)
        return eq_valid and ineq_valid

    def add_usr_defined_eq_constraints(self, A_eq, b_eq):
        """Add user defined constraints.

        The concentrations are restrictred by the equation
        A_eq*c = b_eq

        Parameters:

        A_eq: nested list of numpy array
            NxM matrix where N is the number of equations and
            M is the overall number of concentrations
        b_eq: list or numpy array
            right hand side vector of the equation length N
        """

        A_eq = np.array(A_eq)
        b_eq = np.array(b_eq)
        if len(A_eq.shape) != 2:
            raise InvalidConstraintError("A_eq has to be a 2D matrix!")
        if len(b_eq.shape) != 1:
            raise InvalidConstraintError("b_eq has to be a 1D vector!")

        if A_eq.shape[1] != self.num_concs:
            raise InvalidConstraintError(
                "The number of columns in A_eq has to "
                "match the number of concentration "
                "variables. Hence, A_eq needs to have "
                "{} columns".format(self.num_concs))

        if A_eq.shape[0] != len(b_eq):
            raise InvalidConstraintError(
                "The length of b_eq has to be "
                "the same as the number of rows in A_eq.")

        self.A_eq = np.vstack((self.A_eq, A_eq))
        self.b_eq = np.append(self.b_eq, b_eq)
        self.A_eq, self.b_eq = self._remove_redundant_entries(self.A_eq,
                                                              self.b_eq)
        self._get_interbasis_relations_for_given_matrix(self.A_eq)

    def add_usr_defined_ineq_constraints(self, A_lb, b_lb):
        """Add the user defined constraints.

        The concentrations are restrictred by the equation
        A_lb*c > b_lb

        Parameters:

        A_lb: nested list of numpy array
            NxM matrix where N is the number of equations and
            M is the overall number of concentrations
        b_lb: list or numpy array
            right hand side vector of the equation length N
        """
        A_lb = np.array(A_lb)
        b_lb = np.array(b_lb)
        if len(A_lb.shape) != 2:
            raise InvalidConstraintError("A_lb has to be a 2D matrix!")

        if len(b_lb.shape) != 1:
            raise InvalidConstraintError("b_lb has to be a 1D vector!")

        if A_lb.shape[1] != self.num_concs:
            raise InvalidConstraintError(
                "The number of columns in A_lb has to match the number of "
                "concentration variables. Hence, A_lb needs to have "
                "{} columns".format(self.num_concs))

        if A_lb.shape[0] != len(b_lb):
            raise InvalidConstraintError(
                "The length of b_lb has to be the same as the number of rows "
                "in A_lb.")

        self.A_lb = np.vstack((self.A_lb, A_lb))
        self.b_lb = np.append(self.b_lb, b_lb)
        self.A_lb, self.b_lb = self._remove_redundant_entries(self.A_lb,
                                                              self.b_lb)
        self._get_interbasis_relations_for_given_matrix(self.A_lb)

    def set_conc_ranges(self, ranges):
        """
        Set concentration range based on lower and upper bounds of each
        element.

        Parameters:

        ranges: list
            Nested list of tuples with the same shape as basis_elements.
            If basis_elements is [["Li", "Ru", "X"], ["O", "X"]], ranges
            coulde be [[(0, 1), (0.2, 0.5), (0, 1)], [(0, 0.8), (0, 0.2)]]
        """
        if len(ranges) != len(self.basis_elements):
            raise InvalidConstraintError(
                "The number of bases is wrong. Expected {}, got {}"
                "".format(len(self.basis_elements), len(ranges))
            )

        for item, basis in zip(ranges, self.basis_elements):
            if len(item) != len(basis):
                raise InvalidConstraintError(
                    "Inconsistent number of elements in basis. "
                    "Expected {}, got {}. Basis elements: {}"
                    "".format(len(basis), len(item), self.basis_elements)
                )
            for rng in item:
                if len(rng) != 2:
                    raise InvalidConstraintError(
                        "Provided range needs to be of length 2. "
                        "Lower and upper bound"
                    )
                if any(x < 0.0 or x > 1.0 for x in rng):
                    raise InvalidConstraintError(
                        "Concentrations below 0 or above 1 "
                        "can not be used as concentration constraints"
                    )

        flatten_rng = [item for sublist in ranges for item in sublist]
        A_lb = np.zeros((2*self.num_concs, self.num_concs))
        b_lb = np.zeros(2*self.num_concs)
        for i, item in enumerate(flatten_rng):
            A_lb[2*i, i] = 1
            A_lb[2*i+1, i] = -1
            b_lb[2*i] = item[0]
            b_lb[2*i+1] = -item[1]
        self.add_usr_defined_ineq_constraints(A_lb, b_lb)

    def _get_integers(self, string, variable_range=None):
        """Extract all the integers from a string."""
        if variable_range is None:
            variable_symbols = []
        else:
            variable_symbols = list(variable_range.keys())
        signs = {"+": 1, "-": -1}
        integers = []
        current_indx = 0
        sum_of_variable_coeff = {k: 0 for k in variable_symbols}
        active_sign = 1
        while current_indx < len(string):
            connected_to_symbol = False
            if string[current_indx] in signs.keys():
                active_sign = signs[string[current_indx]]
            elif string[current_indx] == "<":
                active_sign = 1

            if string[current_indx].isdigit():
                substr = string[current_indx]
                # find how many consecutive digits are there
                indx = current_indx+1
                while indx < len(string):
                    # end of integer found
                    if not string[indx].isdigit():
                        # case where integer followed by a variable symbol
                        # (e.g., 4x)
                        if string[indx] in variable_symbols:
                            connected_to_symbol = True
                        break
                    substr += string[indx]
                    indx += 1

                # got an integer number
                if not connected_to_symbol:
                    integers.append(int(substr))
                    current_indx = indx
                else:
                    symbol = string[indx]
                    sum_of_variable_coeff[symbol] += active_sign*int(substr)
                    current_indx = indx + 1

            # case where variable symbol is defined without integer
            elif string[current_indx] in variable_symbols:
                symbol = string[current_indx]
                sum_of_variable_coeff[symbol] += active_sign
                current_indx += 1
            else:
                current_indx += 1

        # Ensure valid chemical formula
        for _, value in sum_of_variable_coeff.items():
            if value != 0:
                raise ValueError("Invalid formula! {} {}"
                                 "".format(string, sum_of_variable_coeff))
        return integers

    def _num_atoms_in_basis(self, formulas, variable_range):
        num_atoms_in_basis = []
        for formula in formulas:
            num_atoms_in_basis.append(
                np.sum(self._get_integers(formula, variable_range)))
        return num_atoms_in_basis

    def set_conc_formula_unit(self, formulas=None, variable_range=None):
        """Set concentration based on formula unit strings.

        Parameters:

        formulas: list
            List constaining formula strings (e.g.,
            ["Li<x>Ru<1>X<2-x>", "O<3-y>X<y>"], ['Al<4-4x>Mg<3x>Si<x>''])
            1. formula string should be provided per basis.
            2. formula string can only have integer numbers.
            3. only one  dvariable is allowed per basis.
            4. each variable should have at least one instance of 'clean'
            representation (e.g., <x>, <y>)

        variable_range: dict
            Range of each variable used in formulas.
            key is a string, and the value should be int or float
            e.g., {"x": (0, 2), "y": (0, 0.7)}, {'x': (0., 1.)}
        """
        if formulas is None or variable_range is None:
            raise InvalidConstraintError("formula and variable range has to"
                                         " be provided!")

        if len(formulas) != len(self.basis_elements):
            raise InvalidConstraintError(
                "Inconsistent number of basis passed. "
                "Expected: {}, got {}. Basis elements: {}"
                "".format(len(self.basis_elements), len(formulas),
                          self.basis_elements)
            )

        element_conc = self._parse_formula_unit_string(formulas)
        num_atoms_in_basis = self._num_atoms_in_basis(formulas, variable_range)

        A_eq = []
        b_eq = []
        col = 0

        num_elements_with_var = self._num_elements_with_var(variable_range,
                                                            element_conc)

        # Set equality matrix and vector due to one or more elements has one
        # fixed concentration value.
        # Note: element_conc is an OrderedDict so k, v pairs come in the
        # same order they were added
        for basis, basis_elem in enumerate(element_conc):
            for k, v in basis_elem.items():
                if v.isdigit():
                    row = np.zeros(self.num_concs)
                    row[col] = num_atoms_in_basis[basis]
                    A_eq.append(row)
                    b_eq.append(int(v))
                col += 1

        variables = list(variable_range.keys())
        # get reference_elements: elements with its concentration specified
        # with a clean representation (e.g., <x> or <y>)
        reference_elements = self._reference_elements(element_conc, variables)

        for variable in variable_range.keys():
            num_additional_constraints = num_elements_with_var[variable] - 2
            if num_additional_constraints == 0:
                continue

            ref_element = reference_elements[variable]["symbol"]
            ref_col = reference_elements[variable]["col"]

            if ref_element is None:
                raise RuntimeError("Did not find reference element for symbol "
                                   "{}".format(variable))

            # Extract the coefficients of other elements with their
            # concenctrations defined as multiples of the symbol
            # (e.g., <2x>, <3x>)
            ignore_conditions = ["+", "-"]
            col = 0
            for basis_elem in element_conc:
                for k, v in basis_elem.items():
                    if k == ref_element or variable not in v:
                        col += 1
                        continue

                    if any(char in v for char in ignore_conditions):
                        col += 1
                        continue
                    coeff = self._get_coeff(v)
                    row = np.zeros(self.num_concs)
                    row[ref_col] = -coeff
                    row[col] = 1
                    A_eq.append(row)
                    b_eq.append(0)
                    col += 1

        if A_eq:
            A_eq = np.array(A_eq)
            self.add_usr_defined_eq_constraints(A_eq, b_eq)

        self._f_u_neq(variable_range, reference_elements, formulas)

    def _num_elements_with_var(self, variable_range, element_conc):
        # count how many elements have their concentration specified with
        # the passed variable.
        num_elements_with_variable = {k: 0 for k in variable_range.keys()}
        for var in variable_range.keys():
            for basis_elem in element_conc:
                for _, conc in basis_elem.items():
                    if var in conc:
                        num_elements_with_variable[var] += 1
        return num_elements_with_variable

    def _f_u_neq(self, variable_range, reference_elements, formulas):
        A_lb = []
        b_lb = []
        num_atoms_in_basis = self._num_atoms_in_basis(formulas, variable_range)
        # For each element in basis
        # figure out if this is reference
        keys = sorted(variable_range.keys())
        for variable in keys:
            rng = variable_range[variable]
            ref_element = reference_elements[variable]["symbol"]
            basis = self._get_basis_containg_variable(formulas, variable)
            col = self._get_col_of_element_in_basis(basis, ref_element)

            # Add lower bound
            row = np.zeros(self.num_concs)
            row[col] = num_atoms_in_basis[basis]
            A_lb.append(row)
            b_lb.append(rng[0])

            # Add upper bound
            row = np.zeros(self.num_concs)
            row[col] = -num_atoms_in_basis[basis]
            A_lb.append(row)
            b_lb.append(-rng[1])

        if A_lb:
            A_lb = np.array(A_lb)
            b_lb = np.array(b_lb)
            self.add_usr_defined_ineq_constraints(A_lb, b_lb)

    def _reference_elements(self, element_conc, variables):
        """Return the reference element for each variable."""
        # reference element is the one that has its concentration specified
        # with a clean representation (e.g., <x> or <y>)
        ref_elem = {}
        for variable in variables:
            col = 0
            for basis_elem in element_conc:
                for k, v in basis_elem.items():
                    if v == variable and v not in ref_elem.keys():
                        ref_elem[variable] = {"symbol": k, "col": col}
                    col += 1
        return ref_elem

    def _get_basis_containg_variable(self, formulas, variable_symbol):
        """Return index of the basis containing the passed varyable symbol."""
        for basis, formula in enumerate(formulas):
            if variable_symbol in formula:
                return basis

        raise ValueError("Did not find any basis containing "
                         "{}".format(variable_symbol))

    def _get_col_of_element_in_basis(self, basis, element):
        """Return column number in the matrix corresponding to element."""
        col = 0
        for i in range(basis):
            col += len(self.basis_elements[i])

        for elem in self.basis_elements[basis]:
            if elem == element:
                return col
            col += 1

        raise RuntimeError("Did not find any column corresponding to "
                           "{} in basis {}. Current basis_elements are {}"
                           "".format(element, basis, self.basis_elements))

    def _get_coeff(self, string):
        """Get the coefficient in front of the symbol.

        Parameters:

        string: str
            string of the following form 3x, 10y, 3z etc.
        """
        if not string[0].isdigit():
            return 1

        number = ""
        for character in string:
            if not character.isdigit():
                break
            number += character
        return int(number)

    def _parse_formula_unit_string(self, formulas):
        element_variable = []
        for basis_num, formula in enumerate(formulas):
            split1 = formula.split("<")
            elements = []
            math = []
            for i, substr in enumerate(split1):
                if i == 0:
                    elements.append(substr)
                elif i == len(split1)-1:
                    split2 = substr.split(">")
                    math.append(split2[0])
                else:
                    split2 = substr.split(">")
                    elements.append(split2[1])
                    math.append(split2[0])
            if elements != self.basis_elements[basis_num]:
                raise ValueError("elements in 'formulas' and 'basis_elements' "
                                 "should match.")
            element_variable.append(OrderedDict(zip(elements, math)))
        return element_variable

    def _get_constraints(self):
        constraints = []
        for i in range(self.A_eq.shape[0]):
            new_constraint = {"type": "eq",
                              "fun": equality_constraint,
                              "args": (self.A_eq[i, :], self.b_eq[i])}
            constraints.append(new_constraint)

        for i in range(self.A_lb.shape[0]):
            new_constraint = {"type": "ineq",
                              "fun": inequality_constraint,
                              "args": (self.A_lb[i, :], self.b_lb[i])}
            constraints.append(new_constraint)
        return constraints

    def _add_fixed_element_in_each_basis(self, nib=None):
        """Add constraints corresponding to the fixing elements in basis."""
        if self.fixed_element_constraint_added:
            return
        from random import choice
        indices = []
        basis_of_index = []
        start = 0
        ranges = self.get_individual_comp_range()
        min_range = 0.01
        maxiter = 1000
        self.orig_num_equality = self.A_eq.shape[0]
        for i, basis in enumerate(self.basis_elements):
            if self._is_linked_to_other_basis(i):
                continue
            iteration = 0
            rng = 0.5*min_range
            while rng < min_range and iteration < maxiter:
                iteration += 1
                indx = choice(range(start, start+len(basis)))
                rng = ranges[indx][1] - ranges[indx][0]
            if iteration >= maxiter:
                self.fixed_element_constraint_added = False
                continue
            indices.append(indx)
            basis_of_index.append(i)
            start += len(basis)

        A = np.zeros((len(indices), self.num_concs))
        b = np.zeros(len(indices))
        for i, indx in enumerate(indices):
            A[i, indx] = 1
            rng = ranges[indx][1] - ranges[indx][0]
            b[i] = np.random.rand()*rng + ranges[indx][0]
            if nib is not None:
                # Convert to closest rational number
                n = nib[basis_of_index[i]]
                b[i] = np.round((b[i]*n))/n

        # Add constraints
        self.A_eq = np.vstack((self.A_eq, A))
        self.b_eq = np.append(self.b_eq, b)
        self.fixed_element_constraint_added = True

    def _remove_fixed_element_in_each_basis_constraint(self):
        """Remove the last rows."""
        if not self.fixed_element_constraint_added:
            return
        self.A_eq = self.A_eq[:self.orig_num_equality, :]
        self.b_eq = self.b_eq[:self.orig_num_equality]
        self.fixed_element_constraint_added = False

    def get_individual_comp_range(self):
        """Return the concentration range of each component."""
        ranges = []
        for i in range(self.num_concs):
            xmin = self.get_conc_min_component(i)
            xmax = self.get_conc_max_component(i)
            ranges.append((xmin[i], xmax[i]))
        return ranges

    def get_random_concentration(self, nib=None):
        """Generate a valid random concentration."""
        assert self.A_eq.shape[0] == len(self.b_eq)

        self._add_fixed_element_in_each_basis(nib=nib)
        # Setup the constraints
        constraints = self._get_constraints()
        x0 = np.random.rand(self.num_concs)

        # Find the closest vector to x0 that satisfies all constraints
        opt_res = minimize(objective_random, x0, args=(x0,),
                           method="SLSQP", jac=obj_jac_random,
                           constraints=constraints,
                           bounds=self.trivial_bounds)
        self._remove_fixed_element_in_each_basis_constraint()
        x = opt_res["x"]

        if not self.is_valid_conc(x):
            msg = "Could not find valid concentration. Revise the constraints."
            raise InvalidConcentrationError(msg)
        return x

    def _get_interbasis_relations(self):
        self._linked_basis = [i for i in range(len(self.basis_elements))]
        self._get_interbasis_relations_for_given_matrix(self.A_eq)
        self._get_interbasis_relations_for_given_matrix(self.A_lb)

    def _get_interbasis_relations_for_given_matrix(self, A):
        basis_start_col = [0]
        for i in range(0, len(self.basis_elements)-1):
            start_col = basis_start_col[-1] + len(self.basis_elements[i])
            basis_start_col.append(start_col)

        # Linked basis
        tol = 1E-6
        for i in range(A.shape[0]):
            lowest_nonzoro_basis = None
            for j, basis_start in enumerate(basis_start_col):
                if j < len(basis_start_col)-1:
                    b_end = basis_start_col[j+1]
                    values = A[i, basis_start:b_end]
                else:
                    values = A[i, basis_start:]

                if np.any(np.abs(values) > tol):
                    # There is a non zero entry in the current basis
                    if lowest_nonzoro_basis is None:
                        lowest_nonzoro_basis = j
                    else:
                        # There exists a link between the two basis
                        root = self._get_root_indx(lowest_nonzoro_basis)
                        self._linked_basis[root] = j

    def _get_root_indx(self, indx):
        """Search find the root index of the link tree."""
        while self._linked_basis[indx] != indx:
            indx = self._linked_basis[indx]
        return indx

    def _is_linked_to_other_basis(self, indx):
        return self._linked_basis[indx] != indx

    @property
    def trivial_bounds(self):
        """Return trivial bounds (i.e. 0 <= x <= 1)

           NOTE: One can give stricter bounds, but these
                 trivial bounds are always satisfied
        """
        return [(0, 1) for _ in range(self.num_concs)]

    def get_conc_min_component(self, comp):
        """Generate all end points of the composition domain."""
        if isinstance(comp, int):
            comp = [comp]
        comp = np.array(comp)
        constraints = self._get_constraints()

        x0 = np.random.rand(self.num_concs)

        # Find the closest vector to x0 that satisfies all constraints
        opt_res = minimize(objective_component_min, x0, args=(comp,),
                           method="SLSQP", jac=obj_jac_component_min,
                           constraints=constraints,
                           bounds=self.trivial_bounds)

        x = opt_res["x"]
        if not self.is_valid_conc(x):
            msg = "Could not find valid concentration. Revise the constraints."
            raise InvalidConcentrationError(msg)
        return x

    def get_conc_max_component(self, comp):
        """Generate all end points of the composition domain."""
        if isinstance(comp, int):
            # Create list with length one
            comp = [comp]
        comp = np.array(comp)
        constraints = self._get_constraints()

        x0 = np.random.rand(self.num_concs)

        # Find the closest vector to x0 that satisfies all constraints
        opt_res = minimize(objective_component_max, x0, args=(comp,),
                           method="SLSQP", jac=obj_jac_component_max,
                           constraints=constraints,
                           bounds=self.trivial_bounds)
        x = opt_res["x"]
        if not self.is_valid_conc(x):
            msg = "Could not find valid concentration. Revise the constraints."
            raise InvalidConcentrationError(msg)
        return x

    def to_float_conc(self, num_atoms_in_basis, int_conc):
        """
        Convert integer number to float concentration.

        Parameters:

        num_atoms_in_basis: list of int
            Number of sites in each basis (e.g., [27, 27], [64]).

        int_conc: array of int
            Concentration per basis in an integer format, which
            corresponds to the number of corresponding element in that basis.
        """
        conc_index = 0
        x = np.zeros(len(int_conc))
        for i, elem in enumerate(self.basis_elements):
            for _ in range(len(elem)):
                x[conc_index] = int_conc[conc_index]/num_atoms_in_basis[i]
                conc_index += 1
        return x

    def conc_in_int(self, num_atoms_in_basis, conc):
        """
        Convert concentration value to an integer that corresponds to the
        number of corresponding elements.

        Parameters:

        num_atoms_in_basis: list of int
            Number of sites in each basis (e.g., [27, 27], [64]).

        conc: array of float
            Concentration per basis normalized to 1.
            (e.g., arrays returned by self.get_random_concentration,
                   self.get_conc_max_component etc.)
        """

        if len(num_atoms_in_basis) != len(self.basis_elements):
            raise ValueError("Number of atoms has to be specified for each "
                             "basis. Given: {}. Expected: {}"
                             "".format(len(num_atoms_in_basis),
                                       len(self.basis_elements)))

        int_array = np.zeros(self.num_concs, dtype=int)
        start = 0
        b_eq = self.b_eq.copy()
        for i, num in enumerate(num_atoms_in_basis):
            n = len(self.basis_elements[i])
            end = start + n
            if end >= len(int_array):
                int_array[start:] = np.round(conc[start:]*num).astype(np.int32)
            else:
                int_array[start: end] = \
                    np.round(conc[start: end]*num).astype(np.int32)
            b_eq[i] *= num
            start += n

        # Make sure that equality constraints are satisfied
        dot_prod = self.A_eq.dot(int_array)
        num_basis = len(num_atoms_in_basis)
        if not np.allclose(dot_prod[:num_basis], b_eq[:num_basis]):
            msg = "The conversion from floating point concentration to int "
            msg += "is not consistent. Expected: {}, ".format(b_eq)
            msg += "got: {}".format(dot_prod)
            raise IntConversionNotConsistentError(msg)
        return int_array

    def _check_grouped_basis_elements(self):
        # check number of basis
        if self.grouped_basis is None:
            return

        num_basis = len([i for sub in self.grouped_basis
                         for i in sub])
        if num_basis != len(self.orig_basis_elements):
            raise ValueError('grouped_basis do not contain all the basis')

        # check if grouped basis have same elements
        for group in self.grouped_basis:
            ref_elements = self.orig_basis_elements[group[0]]
            for indx in group[1:]:
                if self.orig_basis_elements[indx] != ref_elements:
                    raise ValueError('elements in the same group must be same')

    def get_concentration_vector(self, index_by_basis, atoms):
        """Get the concentration vector."""
        assert len(index_by_basis) == len(self.basis_elements)

        concs = np.zeros(self.num_concs)
        start = 0
        for i, indices in enumerate(index_by_basis):
            symbol_lookuptable = \
                {symb: start+j for j, symb
                 in enumerate(self.basis_elements[i])}
            for indx in indices:
                concs[symbol_lookuptable[atoms[indx].symbol]] += 1
            if start+len(self.basis_elements[i]) >= len(concs):
                concs[start:] /= len(indices)
            else:
                concs[start:start+len(self.basis_elements[i])] /= len(indices)
            start += len(self.basis_elements[i])
        assert np.all(concs <= 1.0)
        return concs

    def is_valid(self, index_by_basis, atoms):
        """Check if the atoms object has a valid concentration.

        Parameters:

        index_by_basis: list
            list where the indices of atoms is grouped by basis

        atoms: Atoms object
            wrappend_and_sorted atoms object
        """
        x = self.get_concentration_vector(index_by_basis, atoms)
        return self.is_valid_conc(x)


# Helper function used by the minimization algorithm
def objective_component_min(x, indx):
    return np.sum(x[indx])


def obj_jac_component_min(x, indx):
    jac = np.zeros(len(x))
    jac[indx] = np.sign(x[indx])
    return jac


def objective_component_max(x, indx):
    return -1 * np.sum(x[indx])


def obj_jac_component_max(x, indx):
    jac = np.zeros(len(x))
    jac[indx] = -np.sign(x[indx])
    return jac


def objective_random(x, x0):
    """Return the difference between x and x0."""
    diff = x - x0
    return diff.dot(diff)


def obj_jac_random(x, x0):
    """The jacobian of the objective function."""
    return 2.0*(x-x0)


def equality_constraint(x, vec, rhs):
    """Equality constraint. Return 0 if constraints are satisfied."""
    return vec.dot(x) - rhs


def eq_jac(x, vec, rhs):
    """Jacobian of the equalitu constraint equation."""
    return vec


def inequality_constraint(x, vec, rhs):
    """Inequality constraints. Return a non-negative number if constraints
       are satisfied."""
    return vec.dot(x) - rhs


def ineq_jac(x, vec, rhs):
    """Jacobian of the inequality constraint equations."""
    return vec



from clease.montecarlo.constraints.mc_constraint import MCConstraint
from clease.montecarlo.constraints.constrain_swap_by_basis import ConstrainSwapByBasis
from clease.montecarlo.constraints.constrain_element_insert_by_basis import ConstrainElementInserts
from clease.montecarlo.constraints.fixed_element import FixedElement


from kivy.uix.screenmanager import Screen
from kivy_garden.graph import Graph, LinePlot
from kivy.uix.popup import Popup
from clease.gui.help_message_popup import HelpMessagePopup
from kivy.app import App
from clease.gui.util import parse_concentration_list, parse_temperature_list
from clease.gui.util import parse_comma_sep_list_of_int
from clease.gui.constants import MC_MEAN_CURVE_COLOR
from clease.gui.constants import CONC_PER_BASIS, SYSTEMS_FROM_DB
from clease.gui.mc_runner import MCRunner
from kivy.utils import get_color_from_hex
from threading import Thread
from clease.gui.load_save_dialog import LoadDialog
import json
import os
import traceback


class MCPage(Screen):
    energy_graph = None
    energy_plot = None
    mean_energy_plot = None
    _pop_up = None
    mc_is_running = False

    def on_enter(self):
        self.set_cell_info()
        if self.energy_graph is None:
            self.energy_graph = Graph(
                xlabel='MC sweep',
                ylabel="Mean energy (<E> - <E>[0])",
                x_ticks_minor=0,
                x_ticks_major=10,
                y_ticks_major=10,
                y_grid_label=True,
                x_grid_label=True,
                padding=5,
                xlog=False,
                ylog=False,
                xmin=0.0,
                ymin=0.0,
                precision='%.2f')

            color = get_color_from_hex(MC_MEAN_CURVE_COLOR)
            self.mean_energy_plot = LinePlot(line_width=2, color=color)

            self.energy_graph.add_plot(self.mean_energy_plot)
            self.ids.energyPlot.add_widget(self.energy_graph)

    def dismiss_popup(self):
        self._pop_up.dismiss()
        self._pop_up = None

    def show_temp_help_message(self):
        msg = 'Temperatures can be specified as a comma separated\n'
        msg += 'list of values. The temperatures are given in Kelvin\n'
        msg += 'Example:\n'
        msg += '1000, 900, 800, 700, 600, 500, 400, 300'
        content = HelpMessagePopup(close=self.dismiss_popup, msg=msg)
        self._pop_up = Popup(title="Temperature input", content=content,
                             pos_hint={'right': 0.95, 'top': 0.95},
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def show_conc_help_message(self):
        msg = 'Concentratrions are given as a comma separated list per basis\n'
        msg += 'Example:\n\n'
        msg += 'If basis elements is (Au, Cu, X), (Cu, X)\n'
        msg += 'We can specify concentrations by\n'
        msg += '(0.5, 0.3, 0.2), (0.9, 0.1)\n'
        msg += 'It is important the the concentration sums to one in each '
        msg += 'basis\n\n'
        msg += 'Another option is to give a comma separated list of ids.\n'
        msg += 'The numbers are then interpreted as IDs in the MC database.\n'
        msg += 'To run a sequence of MC calculations, you can therefore \n'
        msg += 'prepare a set of structures and store in the DB.\n'
        msg += 'It is important that the structures match the MC cell.\n'
        msg += 'The easiest way is to click on View MC cell and store it as \n'
        msg += 'xyz file and use that as a template for your simulation\n'
        msg += 'cells.'
        content = HelpMessagePopup(close=self.dismiss_popup, msg=msg)
        self._pop_up = Popup(title="Concentration input", content=content,
                             pos_hint={'right': 0.95, 'top': 0.95},
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def show_size_help_message(self):
        msg = 'Scaling factor used to scale the currently active template\n'
        msg += 'If this factor is 5 a supercell consiting (5, 5, 5)\n'
        msg += 'extension of the active template will be used for the\n'
        msg += 'MC calculation\n'
        content = HelpMessagePopup(close=self.dismiss_popup, msg=msg)
        self._pop_up = Popup(title="Concentration input", content=content,
                             pos_hint={'right': 0.95, 'top': 0.95},
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def to_dict(self):
        return {
            'temps': self.ids.tempInput.text,
            'concs': self.ids.concInput.text,
            'size': self.ids.sizeInput.text,
            'sweeps': self.ids.sweepInput.text,
            'mc_db': self.ids.mc_db_input.text
        }

    def from_dict(self, dct):
        self.ids.tempInput.text = dct.get('temps', '')
        self.ids.concInput.text = dct.get('concs', '')
        self.ids.sizeInput.text = dct.get('size', '1')
        self.ids.sweepInput.text = dct.get('sweeps', '100')
        self.ids.mc_db_input.text = dct.get('mc_db', '')

    def view_mc_cell(self):
        app = App.get_running_app()
        try:
            from ase.visualize import view
            atoms = self._get_mc_cell()
            Thread(target=view, args=(atoms,)).start()
        except Exception as exc:
            app.root.ids.status.text = str(exc)

    def _get_mc_cell(self):
        app = App.get_running_app()
        try:
            settings = app.root.settings
            size = int(self.ids.sizeInput.text)

            if settings is None:
                app.root.ids.status.text = 'Apply settings prior to running MC'
                return

            atoms = settings.atoms*(size, size, size)
            return atoms
        except Exception as exc:
            app.root.ids.status.text = str(exc)
        return None

    def set_cell_info(self):
        atoms = self._get_mc_cell()

        if atoms is None:
            return

        info = atoms.get_cell_lengths_and_angles()

        length_str = 'a: {}Å b: {}Å c: {}Å'.format(
            int(info[0]), int(info[1]), int(info[2]))
        self.ids.mc_cell_lengths.text = length_str

        angle_str = '{}deg {}deg {}deg'.format(int(info[3]), int(info[4]),
                                               int(info[5]))
        self.ids.mc_cell_angles.text = angle_str
        self.ids.mc_num_atoms.text = 'Num atoms: {}'.format(len(atoms))

    def open_load_dialog(self):
        content = LoadDialog(load=self.load_mc_db_file,
                             cancel=self.dismiss_popup)

        self._pop_up = Popup(title="Load MC DB file", content=content,
                             pos_hint={'right': 0.95, 'top': 0.95},
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def load_mc_db_file(self, path, filename):
        if len(filename) == 0:
            fname = path
        else:
            fname = filename[0]

        self.ids.mc_db_input.text = fname
        self.dismiss_popup()

    def show_db_help_message(self):
        msg = 'ASE database used to store thermodynamic data from MC runs\n'
        msg += 'The data are stored in an external table called thermo_data.\n'
        msg += 'The data can thus be accessed from an AtomsRow object by\n'
        msg += "row['thermo_data']"
        content = HelpMessagePopup(close=self.dismiss_popup, msg=msg)
        self._pop_up = Popup(title="MC DB", content=content,
                             pos_hint={'right': 0.95, 'top': 0.95},
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def runMC(self):
        if self.mc_is_running:
            return

        try:
            temps = parse_temperature_list(self.ids.tempInput.text)
            conc_mode = None

            # First try to parse as list
            conc_string = self.ids.concInput.text
            try:
                concs = parse_comma_sep_list_of_int(conc_string)
                conc_mode = SYSTEMS_FROM_DB
            except:
                pass

            if conc_mode is None:
                try:
                    concs = parse_concentration_list(conc_string)
                    conc_mode = CONC_PER_BASIS
                except:
                    pass

            if conc_mode is None:
                msg = 'Cannot parse concentration'
                App.get_running_app().root.ids.status.text = msg
                return

            size = int(self.ids.sizeInput.text)

            app = App.get_running_app()
            settings = app.root.settings

            if settings is None:
                app.root.ids.status.text = 'Apply settings prior to running MC'
                return

            eci_file = app.root.ids.sm.get_screen('Fit').ids.eciFileInput.text
            if not os.path.exists(eci_file):
                msg = 'Cannot load ECI from {}. No such file.'.format(eci_file)
                app.root.ids.status.text = msg
                return

            with open(eci_file, 'r') as infile:
                eci = json.load(infile)

            atoms = settings.atoms*(size, size, size)
            sweeps = int(self.ids.sweepInput.text)
            db_name = self.ids.mc_db_input.text

            runner_args = dict(atoms=atoms, eci=eci, mc_page=self, conc=concs,
                               temps=temps, settings=settings, sweeps=sweeps,
                               status=app.root.ids.status, db_name=db_name,
                               conc_mode=conc_mode)
            if conc_mode == CONC_PER_BASIS:
                runner = MCRunner(**runner_args)
            else:
                # Create a chain of MCRunners
                runner_args['conc'] = None
                runner_args['db_id'] = concs[0]
                runner = MCRunner(**runner_args)
                current_runner = runner
                for i in range(1, len(concs)):
                    runner_args['db_id'] = concs[1]
                    current_runner.next_mc_obj = MCRunner(**runner_args)
                    current_runner = current_runner.next_mc_obj

            Thread(target=runner.run).start()
        except Exception as exc:
            traceback.print_exc()
            App.get_running_app().root.ids.status.text = str(exc)



class MCObserver(object):
    """Base class for all MC observers."""

    def __init__(self):
        self.name = "GenericObserver"

    def __call__(self, system_changes):
        """
        Gets information about the system changes and can perform some action

        Parameters:

        system_changes: list
            List of system changes. For example, if the occupation of the
            atomic index 23 is changed from Mg to Al,
            system_change = [(23, Mg, Al)].
            If an Mg atom occupying the atomic index 26 is swapped with an Al
            atom occupying the atomic index 12,
            system_change = [(26, Mg, Al), (12, Al, Mg)]
        """
        pass

    def reset(self):
        """Reset all values of the MC observer"""
        pass

    def get_averages(self):
        """Return averages in the form of a dictionary."""
        return {}



from setuptools import setup, find_packages, Extension
from distutils.errors import CompileError
from distutils import ccompiler
from textwrap import dedent
from distutils.sysconfig import get_python_inc
from pip._internal import main as pipmain
import os

def src_folder():
    candidates = ['cxx/src/', 'clease/', './']
    for c in candidates:
        if os.path.exists(c+'cluster.cpp'):
            return c
    raise RuntimeError("Cannot find source folder.")

def get_cython_folder():
    candidates = ['cxx/cython/', 'cython/', './']
    for c in candidates:
        if os.path.exists(c+'clease_cxx.pyx'):
            return c
    raise RuntimeError("Cannot find cython folder")

def include_folder():
    candidates = ['cxx/include/', 'clease/', './']
    for c in candidates:
        if os.path.exists(c + 'cluster.hpp'):
            return c
    raise RuntimeError("Cannot find include folder")

def get_npy_include_folder():
    try:
        import numpy
    except ImportError:
        pipmain(['install', 'numpy'])

    import numpy as np
    return np.get_include()

def build_ext(ext_module):
    try:
        import Cython
    except ImportError:
        pipmain(['install', 'cython'])

    from Cython.Build import cythonize

    return cythonize(ext_module)

def install_kivy_garden_from_github():
    try:
        import kivy
    except ImportError:
        pipmain(['install', 'kivy'])

    try:
        from kivy_garden.graph import Graph
    except ImportError:
        pipmain(['install', 'https://github.com/kivy-garden/graph/archive/master.zip'])


cxx_src_folder = src_folder()
cxx_inc_folder = include_folder()
cython_folder = get_cython_folder()
install_kivy_garden_from_github()

src_files = ['cf_history_tracker.cpp',
             'additional_tools.cpp', 'cluster.cpp',
             'row_sparse_struct_matrix.cpp',
             'named_array.cpp', 'symbols_with_numbers.cpp',
             'basis_function.cpp', 'cluster_list.cpp']

src_files = [cxx_src_folder + x for x in src_files]
src_files.append(cython_folder + 'clease_cxx.pyx')
extra_comp_args = ['-std=c++11']

clease_cxx = Extension("clease_cxx", sources=src_files,
                       include_dirs=[cxx_inc_folder, get_npy_include_folder(),
                                     cxx_src_folder, get_python_inc(),
                                     cython_folder],
                       extra_compile_args=extra_comp_args,
                       language="c++")


setup(
    name="clease",
    setup_requires=['cython', 'numpy'],
    ext_modules=build_ext(clease_cxx),
    author=['J. H. Chang', 'D. Kleiven'],
    author_email="jchang@dtu.dk, david.kleiven@ntnu.no",
    long_description='CLuster Expansion in Atomistic Simulation Environment',
    url='https://gitlab.com/computationalmaterials/clease',
    scripts=['bin/clease'],
    version='0.9.5',
    description="CLuster Expansion in Atomistic Simulation Environment",
    packages=find_packages(),
    download_url='https://gitlab.com/computationalmaterials/clease/-/archive/v0.9.5/clease-v0.9.5.zip',
    include_package_data=True,
    package_data={'clease.gui': ['layout/*.kv', 'layout/*.png']},
    license='MPL-2.0',
    keywords=['Cluster Expansion', 'Monte Carlo', 'Computational materials', 'Materials research'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
    install_requires=['ase', 'matplotlib', 'spglib']
)



from clease.montecarlo.observers import MCObserver
import numpy as np


class DiffractionUpdater(object):
    """
    Utility class for all objects that require tracing of Fourier reflection.
    See docstring of DiffractionObserver for explanation of the arguments.
    """

    def __init__(self, atoms=None, k_vector=[], active_symbols=[],
                 all_symbols=[]):
        MCObserver.__init__(self)
        self.orig_symbols = [atom.symbol for atom in atoms]
        self.k_vector = k_vector
        self.N = len(atoms)
        self.k_dot_r = atoms.get_positions().dot(self.k_vector)
        self.indicator = {k: 0 for k in all_symbols}
        for symb in active_symbols:
            self.indicator[symb] = 1.0

        self.value = self.calculate_from_scratch(self.orig_symbols)
        self.prev_value = self.value

    def update(self, system_changes):
        """Update the reflection value."""
        self.prev_value = self.value
        for change in system_changes:
            f_val = np.exp(1j*self.k_dot_r[change[0]])/self.N
            self.value += self.indicator[change[2]]*f_val
            self.value -= self.indicator[change[1]]*f_val

    def undo(self):
        """Undo the last update."""
        self.value = self.prev_value

    def reset(self):
        """Reset all values."""
        self.value = self.calculate_from_scratch(self.orig_symbols)
        self.prev_value = self.value

    def calculate_from_scratch(self, symbols):
        """Calculate the intensity from sctrach."""
        value = 0.0 + 1j*0.0
        for i, symb in enumerate(symbols):
            value += self.indicator[symb]*np.exp(1j*self.k_dot_r[i])
        return value / len(symbols)


class DiffractionObserver(MCObserver):
    """
    Trace the reflection intensity.

    Parameters:

    atoms: Atoms
        Atoms object used in Monte Carlo

    k_vector: list
        Fourier reflection to be traced

    active_symbols: list
        List of symbols that reflects

    all_symbols: list
        List of all symbols in the simulation

    name: str
        Name of the DiffractionObserver
        (users are given the freedom to set names because they can attach
         multiple DiffractionObserver instances)


    Example:

    Consider a system where Al, Mg and Si occupy FCC lattice sites. We want to
    trace the occurence of Mg layers that are separated by a distance 3*a
    where *a* is the lattice parameter. We further assume that the *y*-axis is
    normal to the planes we want to trace. In that case, we specify the
    variables as

    >>> from ase.build import bulk
    >>> import numpy as np
    >>> a = 4.05
    >>> atoms = bulk('Al, crystalstructure='fcc', a=a)
    >>> k_vector = [0, 2.0*np.pi/(3*a), 0]
    >>> active_elements = ['Mg']
    >>> all_symbols = ['Al', 'Mg', 'Si']

    If we do not wish to distinguish Mi and Si (we do not distiguish Mg layer,
    Si layer or a mixture of the two) the `active_elements` is changed to

    >>> active_elements = ['Mg', 'Si']
    """

    def __init__(self, atoms=None, k_vector=[], active_symbols=[],
                 all_symbols=[], name="reflection1"):
        MCObserver.__init__(self)
        self.updater = \
            DiffractionUpdater(atoms=atoms, k_vector=k_vector,
                               active_symbols=active_symbols,
                               all_symbols=all_symbols)
        self.avg = self.updater.value
        self.num_updates = 1
        self.name = name

    def __call__(self, system_changes):
        self.updater.update(system_changes)
        self.avg += self.updater.value

    def get_averages(self):
        return {self.name: np.abs(self.avg/self.num_updates)}

    def reset(self):
        self.updater.reset()
        self.avg = self.updater.value
        self.num_updates = 1



from clease.cleaselogger import CLEASELogger, LogVerbosity
_logger = CLEASELogger()


def set_verbosity(verbosity):
    _logger.verbosity = verbosity


def set_fd(fd):
    _logger.fd = fd

from clease.template_filters import SkewnessFilter, EquivalentCellsFilter
from clease.template_filters import ValidConcentrationFilter
from clease.template_filters import DistanceBetweenFacetsFilter
from clease.template_filters import VolumeToSurfaceRatioFilter
from clease.template_filters import AngleFilter
from clease.atoms_manager import AtomsManager
from clease.cluster_extractor import ClusterExtractor
from clease.settings_bulk import CEBulk, CECrystal
from clease.evaluate import Evaluate
from clease.corrFunc import CorrFunction
from clease.newStruct import NewStructures
from clease.convexhull import ConvexHull
from clease.concentration import Concentration
from clease.regression import LinearRegression, Tikhonov, Lasso
from clease.ga_fit import GAFit
from clease.bayesian_compressive_sensing import BayesianCompressiveSensing
from clease.cluster_info_mapper import ClusterInfoMapper

__all__ = ['CEBulk', 'CECrystal', 'Concentration', 'CorrFunction',
           'NewStructures', 'NewStructures', 'Evaluate', 'AtomsManager',
           'ConvexHull', 'LinearRegression', 'Tikhonov', 'Lasso',
           'GAFit', 'BayesianCompressiveSensing', 'LogVerbosity',
           'SkewnessFilter', 'ClusterExtractor']
